# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from opentelemetry.proto.collector.logs.v1 import (
    logs_service_pb2 as opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2,
)

GRPC_GENERATED_VERSION = "1.63.2"
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = "1.65.0"
SCHEDULED_RELEASE_DATE = "June 25, 2024"
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower

    _version_not_supported = first_version_is_lower(
        GRPC_VERSION, GRPC_GENERATED_VERSION
    )
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f"The grpc package installed is at version {GRPC_VERSION},"
        + f" but the generated code in opentelemetry/proto/collector/logs/v1/logs_service_pb2_grpc.py depends on"
        + f" grpcio>={GRPC_GENERATED_VERSION}."
        + f" Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}"
        + f" or downgrade your generated code using grpcio-tools<={GRPC_VERSION}."
        + f" This warning will become an error in {EXPECTED_ERROR_RELEASE},"
        + f" scheduled for release on {SCHEDULED_RELEASE_DATE}.",
        RuntimeWarning,
    )


class LogsServiceStub(object):
    """Service that can be used to push logs between one Application instrumented with
    OpenTelemetry and an collector, or between an collector and a central collector (in this
    case logs are sent/received to/from multiple Applications).
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Export = channel.unary_unary(
            "/opentelemetry.proto.collector.logs.v1.LogsService/Export",
            request_serializer=opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2.ExportLogsServiceRequest.SerializeToString,
            response_deserializer=opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2.ExportLogsServiceResponse.FromString,
            _registered_method=True,
        )


class LogsServiceServicer(object):
    """Service that can be used to push logs between one Application instrumented with
    OpenTelemetry and an collector, or between an collector and a central collector (in this
    case logs are sent/received to/from multiple Applications).
    """

    def Export(self, request, context):
        """For performance reasons, it is recommended to keep this RPC
        alive for the entire life of the application.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_LogsServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Export": grpc.unary_unary_rpc_method_handler(
            servicer.Export,
            request_deserializer=opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2.ExportLogsServiceRequest.FromString,
            response_serializer=opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2.ExportLogsServiceResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "opentelemetry.proto.collector.logs.v1.LogsService",
        rpc_method_handlers,
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class LogsService(object):
    """Service that can be used to push logs between one Application instrumented with
    OpenTelemetry and an collector, or between an collector and a central collector (in this
    case logs are sent/received to/from multiple Applications).
    """

    @staticmethod
    def Export(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/opentelemetry.proto.collector.logs.v1.LogsService/Export",
            opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2.ExportLogsServiceRequest.SerializeToString,
            opentelemetry_dot_proto_dot_collector_dot_logs_dot_v1_dot_logs__service__pb2.ExportLogsServiceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )



# iva.py - functions for handling Italian VAT numbers
#
# Copyright (C) 2012, 2013 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Partita IVA (Italian VAT number).

The Partita IVA (Imposta sul valore aggiunto) consists of 11 digits. The
first 7 digits are a company identifier, the next 3 refer to the province
of residence and the last is a check digit.

The fiscal code for individuals is not accepted as valid code for
intracommunity VAT related operations so it is ignored here.

>>> validate('IT 00743110157')
'00743110157'
>>> validate('00743110158')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number, ' -:').upper().strip()
    if number.startswith('IT'):
        number = number[2:]
    return number


def validate(number):
    """Check if the number is a valid VAT number. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if not isdigits(number) or int(number[0:7]) == 0:
        raise InvalidFormat()
    if len(number) != 11:
        raise InvalidLength()
    # check the province of residence
    if not ('001' <= number[7:10] <= '100' or number[7:10] in ('120', '121', '888', '999')):
        raise InvalidComponent()
    luhn.validate(number)
    return number


def is_valid(number):
    """Check if the number is a valid VAT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from enum import IntEnum

from messagebus import TokenCreator, TokenPriority


MSGGRP_SLCTRL_PRODMGR = "9"


def _create_product_token(number):
    return TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group=MSGGRP_SLCTRL_PRODMGR,
        number=number,
    )


MSGGRP_SLCTRL_RUPTURE = "42"


def _create_rupture_token(number):
    return TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group=MSGGRP_SLCTRL_RUPTURE,
        number=number,
    )

# noinspection SpellCheckingInspection


class ProductApiTokens(IntEnum):

    TK_SLCTRL_PMGR_GET_TENDERS = _create_product_token("68")
    TK_SLCTRL_PMGR_GET_PRODUCTS_NAME = _create_product_token("69")
    TK_SLCTRL_PMGR_RUPTURE_UPDATE_DAILY_DISHES = _create_rupture_token("13")
    TK_SLCTRL_PMGR_RUPTURE_UPDATE_ITEMS = _create_rupture_token("8")



from datetime import datetime
from xml.etree.ElementTree import Element

import iso8601
from ._BaseLoyaltyCoupon import BaseLoyaltyCoupon


class LoyaltyCouponNfceBuilder(BaseLoyaltyCoupon):


    def get_fiscal_key(self, fiscal_xml: Element) -> str:
        return fiscal_xml.find(".//infNFe").attrib["Id"][3:]

    def get_fiscal_date(self, fiscal_xml: Element) -> datetime:
        return iso8601.parse_date(fiscal_xml.find(".//dhEmi").text)



import logging

from model.customexception import ProductNotExistsException

from actions.do_complete_option import doCompleteOption
from actions.products.get_products import (
    get_product_by_barcode,
    get_product_price_from_cache,
)
from api.order import get_order_service
from order_api.dto.exception import CannotAddProductWithMandatoryCustomizationOnTotal
from sysactions import (
    action,
    show_messagebox,
)

logger = logging.getLogger("PosActions")


@action
def do_sell_product_by_barcode(
    pos_id,                     # type: str
    barcode_number,             # type: str
    check_consistency=False,    # type: bool
    received_qty="1",           # type: str
    seat="",                    # type: str
):
    # type: (...) -> str
    return sell_product_by_barcode(
        pos_id=pos_id,
        barcode_number=barcode_number,
        check_consistency=check_consistency,
        received_qty=received_qty,
        seat=seat,
    )


def sell_product_by_barcode(
    pos_id,                     # type: str
    barcode_number,             # type: str
    check_consistency=False,    # type: bool
    received_qty="1",           # type: str
    seat="",                    # type: str
):
    # type: (...) -> str

    try:
        if not barcode_number:
            return "False"

        barcode_number = str(barcode_number).zfill(13)
        if check_consistency and not _is_a_valid_ean13(barcode=barcode_number):
            logger.info("Barcode {0} is not a Valid EAN13".format(barcode_number))
            show_messagebox(
                posid=pos_id,
                message="$BARCODE_WRONG",
                buttons="$OK",
            )
            return "False"

        success = _sell_product(
            pos_id=pos_id,
            barcode_number=str(barcode_number),
            quantity=received_qty,
            seat=seat,
        )

        if not success:
            return "False"

        order_service = get_order_service(pos_id=pos_id)
        order_picture = order_service.get_order_picture()
        if order_picture is not None and order_picture.get("state") == "TOTALED":
            show_messagebox(
                posid=pos_id,
                message="$PRODUCT_SUCCESSFULLY_ADDED",
                timeout=500,
            )

        table_id_prop = order_picture.find(".//CustomOrderProperties/OrderProperty[@key='{}']".format("TABLE_ID"))
        if table_id_prop is not None:
            from msgbusboundary import MsgBusTableService as tableService
            from table_actions import create_table_modified_event

            table_id = table_id_prop.get("value", None)
            table_picture = tableService().get_table_picture(pos_id, table_id)
            create_table_modified_event(pos_id, table_picture)

        return "True"
    except ProductNotExistsException:
        return "Clear"
    except CannotAddProductWithMandatoryCustomizationOnTotal as ex:
        show_messagebox(
            posid=pos_id,
            message="$PRODUCT_WITH_MANDATORY_CUSTOMIZATIONS_ON_TOTAL_ORDER_ERROR|{}".format(ex.part_name),
        )
    except (Exception,):
        logger.exception("Error on [sell_product_by_barcode]")
        return "False"


def _sell_product(
    pos_id,             # typs: str
    barcode_number,     # type: str
    quantity,           # type: str
    seat,               # type: str
):
    # type: (...) -> bool

    if barcode_number.startswith("2"):
        logger.info("Barcode {0} is a Scale Type Code".format(barcode_number))
        plu, calc_qty = _ticket_from_scale(
            barcode=barcode_number,
            pos_id=pos_id,
        )
        doCompleteOption(
            pos_id=pos_id,
            context="1",
            part_code=plu,
            qty=calc_qty,
            seat=seat,
        )

        return True

    logger.info("Barcode {0} ".format(barcode_number))
    product_by_barcode = get_product_by_barcode(barcode=barcode_number)

    if product_by_barcode:
        response = doCompleteOption(
            pos_id=pos_id,
            context="1",
            part_code=product_by_barcode["plu"],
            qty=quantity,
            seat=seat,
            barcode_number=product_by_barcode["name"],
        )
        if response == "Clear":
            return False

        return True

    logger.info("Scanned barcode {0} not found on the product database!".format(barcode_number))
    show_messagebox(
        posid=pos_id,
        message="$BARCODE_NOT_FOUND|{}".format(barcode_number),
        buttons="$OK",
    )

    return False


def _is_a_valid_ean13(
    barcode,   # type: str
):
    # type: (...) -> bool

    if len(str(barcode)) > 13:
        return False

    barcode = list(barcode)
    index = 0
    value = 0
    dv = barcode.pop()
    for x in barcode:
        value = value + (int(x) * ((index % 2) * 2 + 1))
        index += 1
    value2 = (((int(value / 10) + 1) * 10) - value) % 10
    if value2 != int(dv):
        return False
    return True


def _ticket_from_scale(
    barcode,   # type: str
    pos_id,    # type: str
):
    # type: (...) -> (str, float)

    from posactions import (
        scale_ticket_mode,
        barcode_pattern_reading,
    )

    try:
        first_x = barcode_pattern_reading.find("X")
        first_0 = barcode_pattern_reading.find("0")
        first_y = barcode_pattern_reading.find("Y")
        first_w = barcode_pattern_reading.find("W")
        plu = int(barcode[first_x:first_0 if first_0 > -1 else first_y])
        param = int(barcode[first_y:first_w])

        if scale_ticket_mode == 'price':
            final_price = float(param) / 100
            price = float(get_product_price_from_cache(plu, pos_id))
            qtd = round((final_price / price), 3)
        else:
            qtd = round(float(param) / 1000, 3)

        return str(plu), qtd
    except ProductNotExistsException as ex:
        logger.info("The product {} in Barcode does not exists".format(ex.message))
        show_messagebox(pos_id, "$PRODUCT_DOES_NOT_EXISTS", "$ERROR")
        raise



# -*- coding: utf-8 -*-
"""
Tests for greenlet.

"""
import os
import sys
import unittest

from gc import collect
from gc import get_objects
from threading import active_count as active_thread_count
from time import sleep
from time import time

import psutil

from greenlet import greenlet as RawGreenlet
from greenlet import getcurrent

from greenlet._greenlet import get_pending_cleanup_count
from greenlet._greenlet import get_total_main_greenlets

from . import leakcheck

PY312 = sys.version_info[:2] >= (3, 12)
PY313 = sys.version_info[:2] >= (3, 13)

WIN = sys.platform.startswith("win")
RUNNING_ON_GITHUB_ACTIONS = os.environ.get('GITHUB_ACTIONS')
RUNNING_ON_TRAVIS = os.environ.get('TRAVIS') or RUNNING_ON_GITHUB_ACTIONS
RUNNING_ON_APPVEYOR = os.environ.get('APPVEYOR')
RUNNING_ON_CI = RUNNING_ON_TRAVIS or RUNNING_ON_APPVEYOR
RUNNING_ON_MANYLINUX = os.environ.get('GREENLET_MANYLINUX')

class TestCaseMetaClass(type):
    # wrap each test method with
    # a) leak checks
    def __new__(cls, classname, bases, classDict):
        # pylint and pep8 fight over what this should be called (mcs or cls).
        # pylint gets it right, but we can't scope disable pep8, so we go with
        # its convention.
        # pylint: disable=bad-mcs-classmethod-argument
        check_totalrefcount = True

        # Python 3: must copy, we mutate the classDict. Interestingly enough,
        # it doesn't actually error out, but under 3.6 we wind up wrapping
        # and re-wrapping the same items over and over and over.
        for key, value in list(classDict.items()):
            if key.startswith('test') and callable(value):
                classDict.pop(key)
                if check_totalrefcount:
                    value = leakcheck.wrap_refcount(value)
                classDict[key] = value
        return type.__new__(cls, classname, bases, classDict)


class TestCase(TestCaseMetaClass(
        "NewBase",
        (unittest.TestCase,),
        {})):

    cleanup_attempt_sleep_duration = 0.001
    cleanup_max_sleep_seconds = 1

    def wait_for_pending_cleanups(self,
                                  initial_active_threads=None,
                                  initial_main_greenlets=None):
        initial_active_threads = initial_active_threads or self.threads_before_test
        initial_main_greenlets = initial_main_greenlets or self.main_greenlets_before_test
        sleep_time = self.cleanup_attempt_sleep_duration
        # NOTE: This is racy! A Python-level thread object may be dead
        # and gone, but the C thread may not yet have fired its
        # destructors and added to the queue. There's no particular
        # way to know that's about to happen. We try to watch the
        # Python threads to make sure they, at least, have gone away.
        # Counting the main greenlets, which we can easily do deterministically,
        # also helps.

        # Always sleep at least once to let other threads run
        sleep(sleep_time)
        quit_after = time() + self.cleanup_max_sleep_seconds
        # TODO: We could add an API that calls us back when a particular main greenlet is deleted?
        # It would have to drop the GIL
        while (
                get_pending_cleanup_count()
                or active_thread_count() > initial_active_threads
                or (not self.expect_greenlet_leak
                    and get_total_main_greenlets() > initial_main_greenlets)):
            sleep(sleep_time)
            if time() > quit_after:
                print("Time limit exceeded.")
                print("Threads: Waiting for only", initial_active_threads,
                      "-->", active_thread_count())
                print("MGlets : Waiting for only", initial_main_greenlets,
                      "-->", get_total_main_greenlets())
                break
        collect()

    def count_objects(self, kind=list, exact_kind=True):
        # pylint:disable=unidiomatic-typecheck
        # Collect the garbage.
        for _ in range(3):
            collect()
        if exact_kind:
            return sum(
                1
                for x in get_objects()
                if type(x) is kind
            )
        # instances
        return sum(
            1
            for x in get_objects()
            if isinstance(x, kind)
        )

    greenlets_before_test = 0
    threads_before_test = 0
    main_greenlets_before_test = 0
    expect_greenlet_leak = False

    def count_greenlets(self):
        """
        Find all the greenlets and subclasses tracked by the GC.
        """
        return self.count_objects(RawGreenlet, False)

    def setUp(self):
        # Ensure the main greenlet exists, otherwise the first test
        # gets a false positive leak
        super().setUp()
        getcurrent()
        self.threads_before_test = active_thread_count()
        self.main_greenlets_before_test = get_total_main_greenlets()
        self.wait_for_pending_cleanups(self.threads_before_test, self.main_greenlets_before_test)
        self.greenlets_before_test = self.count_greenlets()

    def tearDown(self):
        if getattr(self, 'skipTearDown', False):
            return

        self.wait_for_pending_cleanups(self.threads_before_test, self.main_greenlets_before_test)
        super().tearDown()

    def get_expected_returncodes_for_aborted_process(self):
        import signal
        # The child should be aborted in an unusual way. On POSIX
        # platforms, this is done with abort() and signal.SIGABRT,
        # which is reflected in a negative return value; however, on
        # Windows, even though we observe the child print "Fatal
        # Python error: Aborted" and in older versions of the C
        # runtime "This application has requested the Runtime to
        # terminate it in an unusual way," it always has an exit code
        # of 3. This is interesting because 3 is the error code for
        # ERROR_PATH_NOT_FOUND; BUT: the C runtime abort() function
        # also uses this code.
        #
        # If we link to the static C library on Windows, the error
        # code changes to '0xc0000409' (hex(3221226505)), which
        # apparently is STATUS_STACK_BUFFER_OVERRUN; but "What this
        # means is that nowadays when you get a
        # STATUS_STACK_BUFFER_OVERRUN, it doesn’t actually mean that
        # there is a stack buffer overrun. It just means that the
        # application decided to terminate itself with great haste."
        #
        #
        # On windows, we've also seen '0xc0000005' (hex(3221225477)).
        # That's "Access Violation"
        #
        # See
        # https://devblogs.microsoft.com/oldnewthing/20110519-00/?p=10623
        # and
        # https://docs.microsoft.com/en-us/previous-versions/k089yyh0(v=vs.140)?redirectedfrom=MSDN
        # and
        # https://devblogs.microsoft.com/oldnewthing/20190108-00/?p=100655
        expected_exit = (
            -signal.SIGABRT,
            # But beginning on Python 3.11, the faulthandler
            # that prints the C backtraces sometimes segfaults after
            # reporting the exception but before printing the stack.
            # This has only been seen on linux/gcc.
            -signal.SIGSEGV,
        ) if not WIN else (
            3,
            0xc0000409,
            0xc0000005,
        )
        return expected_exit

    def get_process_uss(self):
        """
        Return the current process's USS in bytes.

        uss is available on Linux, macOS, Windows. Also known as
        "Unique Set Size", this is the memory which is unique to a
        process and which would be freed if the process was terminated
        right now.

        If this is not supported by ``psutil``, this raises the
        :exc:`unittest.SkipTest` exception.
        """
        try:
            return psutil.Process().memory_full_info().uss
        except AttributeError as e:
            raise unittest.SkipTest("uss not supported") from e

    def run_script(self, script_name, show_output=True):
        import subprocess
        script = os.path.join(
            os.path.dirname(__file__),
            script_name,
        )

        try:
            return subprocess.check_output([sys.executable, script],
                                           encoding='utf-8',
                                           stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as ex:
            if show_output:
                print('-----')
                print('Failed to run script', script)
                print('~~~~~')
                print(ex.output)
                print('------')
            raise


    def assertScriptRaises(self, script_name, exitcodes=None):
        import subprocess
        with self.assertRaises(subprocess.CalledProcessError) as exc:
            output = self.run_script(script_name, show_output=False)
            __traceback_info__ = output
            # We're going to fail the assertion if we get here, at least
            # preserve the output in the traceback.

        if exitcodes is None:
            exitcodes = self.get_expected_returncodes_for_aborted_process()
        self.assertIn(exc.exception.returncode, exitcodes)
        return exc.exception



from sqlalchemy import Column
from sqlalchemy import ForeignKeyConstraint
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Table

from ._autogen_fixtures import AutogenFixtureTest
from ...testing import combinations
from ...testing import config
from ...testing import eq_
from ...testing import mock
from ...testing import TestBase


class AutogenerateForeignKeysTest(AutogenFixtureTest, TestBase):
    __backend__ = True
    __requires__ = ("foreign_key_constraint_reflection",)

    def test_remove_fk(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "some_table",
            m1,
            Column("test", String(10), primary_key=True),
        )

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("test2", String(10)),
            ForeignKeyConstraint(["test2"], ["some_table.test"]),
        )

        Table(
            "some_table",
            m2,
            Column("test", String(10), primary_key=True),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("test2", String(10)),
        )

        diffs = self._fixture(m1, m2)

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["test2"],
            "some_table",
            ["test"],
            conditional_name="servergenerated",
        )

    def test_add_fk(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "some_table",
            m1,
            Column("id", Integer, primary_key=True),
            Column("test", String(10)),
        )

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("test2", String(10)),
        )

        Table(
            "some_table",
            m2,
            Column("id", Integer, primary_key=True),
            Column("test", String(10)),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("test2", String(10)),
            ForeignKeyConstraint(["test2"], ["some_table.test"]),
        )

        diffs = self._fixture(m1, m2)

        self._assert_fk_diff(
            diffs[0], "add_fk", "user", ["test2"], "some_table", ["test"]
        )

    def test_no_change(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "some_table",
            m1,
            Column("id", Integer, primary_key=True),
            Column("test", String(10)),
        )

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("test2", Integer),
            ForeignKeyConstraint(["test2"], ["some_table.id"]),
        )

        Table(
            "some_table",
            m2,
            Column("id", Integer, primary_key=True),
            Column("test", String(10)),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("test2", Integer),
            ForeignKeyConstraint(["test2"], ["some_table.id"]),
        )

        diffs = self._fixture(m1, m2)

        eq_(diffs, [])

    def test_no_change_composite_fk(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "some_table",
            m1,
            Column("id_1", String(10), primary_key=True),
            Column("id_2", String(10), primary_key=True),
        )

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("other_id_1", String(10)),
            Column("other_id_2", String(10)),
            ForeignKeyConstraint(
                ["other_id_1", "other_id_2"],
                ["some_table.id_1", "some_table.id_2"],
            ),
        )

        Table(
            "some_table",
            m2,
            Column("id_1", String(10), primary_key=True),
            Column("id_2", String(10), primary_key=True),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("other_id_1", String(10)),
            Column("other_id_2", String(10)),
            ForeignKeyConstraint(
                ["other_id_1", "other_id_2"],
                ["some_table.id_1", "some_table.id_2"],
            ),
        )

        diffs = self._fixture(m1, m2)

        eq_(diffs, [])

    def test_casing_convention_changed_so_put_drops_first(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "some_table",
            m1,
            Column("test", String(10), primary_key=True),
        )

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("test2", String(10)),
            ForeignKeyConstraint(["test2"], ["some_table.test"], name="MyFK"),
        )

        Table(
            "some_table",
            m2,
            Column("test", String(10), primary_key=True),
        )

        # foreign key autogen currently does not take "name" into account,
        # so change the def just for the purposes of testing the
        # add/drop order for now.
        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("test2", String(10)),
            ForeignKeyConstraint(["a1"], ["some_table.test"], name="myfk"),
        )

        diffs = self._fixture(m1, m2)

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["test2"],
            "some_table",
            ["test"],
            name="MyFK" if config.requirements.fk_names.enabled else None,
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["a1"],
            "some_table",
            ["test"],
            name="myfk",
        )

    def test_add_composite_fk_with_name(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "some_table",
            m1,
            Column("id_1", String(10), primary_key=True),
            Column("id_2", String(10), primary_key=True),
        )

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("other_id_1", String(10)),
            Column("other_id_2", String(10)),
        )

        Table(
            "some_table",
            m2,
            Column("id_1", String(10), primary_key=True),
            Column("id_2", String(10), primary_key=True),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("other_id_1", String(10)),
            Column("other_id_2", String(10)),
            ForeignKeyConstraint(
                ["other_id_1", "other_id_2"],
                ["some_table.id_1", "some_table.id_2"],
                name="fk_test_name",
            ),
        )

        diffs = self._fixture(m1, m2)
        self._assert_fk_diff(
            diffs[0],
            "add_fk",
            "user",
            ["other_id_1", "other_id_2"],
            "some_table",
            ["id_1", "id_2"],
            name="fk_test_name",
        )

    @config.requirements.no_name_normalize
    def test_remove_composite_fk(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "some_table",
            m1,
            Column("id_1", String(10), primary_key=True),
            Column("id_2", String(10), primary_key=True),
        )

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("other_id_1", String(10)),
            Column("other_id_2", String(10)),
            ForeignKeyConstraint(
                ["other_id_1", "other_id_2"],
                ["some_table.id_1", "some_table.id_2"],
                name="fk_test_name",
            ),
        )

        Table(
            "some_table",
            m2,
            Column("id_1", String(10), primary_key=True),
            Column("id_2", String(10), primary_key=True),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("a1", String(10), server_default="x"),
            Column("other_id_1", String(10)),
            Column("other_id_2", String(10)),
        )

        diffs = self._fixture(m1, m2)

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["other_id_1", "other_id_2"],
            "some_table",
            ["id_1", "id_2"],
            conditional_name="fk_test_name",
        )

    def test_add_fk_colkeys(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "some_table",
            m1,
            Column("id_1", String(10), primary_key=True),
            Column("id_2", String(10), primary_key=True),
        )

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("other_id_1", String(10)),
            Column("other_id_2", String(10)),
        )

        Table(
            "some_table",
            m2,
            Column("id_1", String(10), key="tid1", primary_key=True),
            Column("id_2", String(10), key="tid2", primary_key=True),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("other_id_1", String(10), key="oid1"),
            Column("other_id_2", String(10), key="oid2"),
            ForeignKeyConstraint(
                ["oid1", "oid2"],
                ["some_table.tid1", "some_table.tid2"],
                name="fk_test_name",
            ),
        )

        diffs = self._fixture(m1, m2)

        self._assert_fk_diff(
            diffs[0],
            "add_fk",
            "user",
            ["other_id_1", "other_id_2"],
            "some_table",
            ["id_1", "id_2"],
            name="fk_test_name",
        )

    def test_no_change_colkeys(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "some_table",
            m1,
            Column("id_1", String(10), primary_key=True),
            Column("id_2", String(10), primary_key=True),
        )

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("other_id_1", String(10)),
            Column("other_id_2", String(10)),
            ForeignKeyConstraint(
                ["other_id_1", "other_id_2"],
                ["some_table.id_1", "some_table.id_2"],
            ),
        )

        Table(
            "some_table",
            m2,
            Column("id_1", String(10), key="tid1", primary_key=True),
            Column("id_2", String(10), key="tid2", primary_key=True),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("other_id_1", String(10), key="oid1"),
            Column("other_id_2", String(10), key="oid2"),
            ForeignKeyConstraint(
                ["oid1", "oid2"], ["some_table.tid1", "some_table.tid2"]
            ),
        )

        diffs = self._fixture(m1, m2)

        eq_(diffs, [])


class IncludeHooksTest(AutogenFixtureTest, TestBase):
    __backend__ = True
    __requires__ = ("fk_names",)

    @combinations(("object",), ("name",))
    @config.requirements.no_name_normalize
    def test_remove_connection_fk(self, hook_type):
        m1 = MetaData()
        m2 = MetaData()

        ref = Table(
            "ref",
            m1,
            Column("id", Integer, primary_key=True),
        )
        t1 = Table(
            "t",
            m1,
            Column("x", Integer),
            Column("y", Integer),
        )
        t1.append_constraint(
            ForeignKeyConstraint([t1.c.x], [ref.c.id], name="fk1")
        )
        t1.append_constraint(
            ForeignKeyConstraint([t1.c.y], [ref.c.id], name="fk2")
        )

        ref = Table(
            "ref",
            m2,
            Column("id", Integer, primary_key=True),
        )
        Table(
            "t",
            m2,
            Column("x", Integer),
            Column("y", Integer),
        )

        if hook_type == "object":

            def include_object(object_, name, type_, reflected, compare_to):
                return not (
                    isinstance(object_, ForeignKeyConstraint)
                    and type_ == "foreign_key_constraint"
                    and reflected
                    and name == "fk1"
                )

            diffs = self._fixture(m1, m2, object_filters=include_object)
        elif hook_type == "name":

            def include_name(name, type_, parent_names):
                if name == "fk1":
                    if type_ == "index":  # MariaDB thing
                        return True
                    eq_(type_, "foreign_key_constraint")
                    eq_(
                        parent_names,
                        {
                            "schema_name": None,
                            "table_name": "t",
                            "schema_qualified_table_name": "t",
                        },
                    )
                    return False
                else:
                    return True

            diffs = self._fixture(m1, m2, name_filters=include_name)

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "t",
            ["y"],
            "ref",
            ["id"],
            conditional_name="fk2",
        )
        eq_(len(diffs), 1)

    def test_add_metadata_fk(self):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "ref",
            m1,
            Column("id", Integer, primary_key=True),
        )
        Table(
            "t",
            m1,
            Column("x", Integer),
            Column("y", Integer),
        )

        ref = Table(
            "ref",
            m2,
            Column("id", Integer, primary_key=True),
        )
        t2 = Table(
            "t",
            m2,
            Column("x", Integer),
            Column("y", Integer),
        )
        t2.append_constraint(
            ForeignKeyConstraint([t2.c.x], [ref.c.id], name="fk1")
        )
        t2.append_constraint(
            ForeignKeyConstraint([t2.c.y], [ref.c.id], name="fk2")
        )

        def include_object(object_, name, type_, reflected, compare_to):
            return not (
                isinstance(object_, ForeignKeyConstraint)
                and type_ == "foreign_key_constraint"
                and not reflected
                and name == "fk1"
            )

        diffs = self._fixture(m1, m2, object_filters=include_object)

        self._assert_fk_diff(
            diffs[0], "add_fk", "t", ["y"], "ref", ["id"], name="fk2"
        )
        eq_(len(diffs), 1)

    @combinations(("object",), ("name",))
    @config.requirements.no_name_normalize
    def test_change_fk(self, hook_type):
        m1 = MetaData()
        m2 = MetaData()

        r1a = Table(
            "ref_a",
            m1,
            Column("a", Integer, primary_key=True),
        )
        Table(
            "ref_b",
            m1,
            Column("a", Integer, primary_key=True),
            Column("b", Integer, primary_key=True),
        )
        t1 = Table(
            "t",
            m1,
            Column("x", Integer),
            Column("y", Integer),
            Column("z", Integer),
        )
        t1.append_constraint(
            ForeignKeyConstraint([t1.c.x], [r1a.c.a], name="fk1")
        )
        t1.append_constraint(
            ForeignKeyConstraint([t1.c.y], [r1a.c.a], name="fk2")
        )

        Table(
            "ref_a",
            m2,
            Column("a", Integer, primary_key=True),
        )
        r2b = Table(
            "ref_b",
            m2,
            Column("a", Integer, primary_key=True),
            Column("b", Integer, primary_key=True),
        )
        t2 = Table(
            "t",
            m2,
            Column("x", Integer),
            Column("y", Integer),
            Column("z", Integer),
        )
        t2.append_constraint(
            ForeignKeyConstraint(
                [t2.c.x, t2.c.z], [r2b.c.a, r2b.c.b], name="fk1"
            )
        )
        t2.append_constraint(
            ForeignKeyConstraint(
                [t2.c.y, t2.c.z], [r2b.c.a, r2b.c.b], name="fk2"
            )
        )

        if hook_type == "object":

            def include_object(object_, name, type_, reflected, compare_to):
                return not (
                    isinstance(object_, ForeignKeyConstraint)
                    and type_ == "foreign_key_constraint"
                    and name == "fk1"
                )

            diffs = self._fixture(m1, m2, object_filters=include_object)
        elif hook_type == "name":

            def include_name(name, type_, parent_names):
                if type_ == "index":
                    return True  # MariaDB thing

                if name == "fk1":
                    eq_(type_, "foreign_key_constraint")
                    eq_(
                        parent_names,
                        {
                            "schema_name": None,
                            "table_name": "t",
                            "schema_qualified_table_name": "t",
                        },
                    )
                    return False
                else:
                    return True

            diffs = self._fixture(m1, m2, name_filters=include_name)

        if hook_type == "object":
            self._assert_fk_diff(
                diffs[0], "remove_fk", "t", ["y"], "ref_a", ["a"], name="fk2"
            )
            self._assert_fk_diff(
                diffs[1],
                "add_fk",
                "t",
                ["y", "z"],
                "ref_b",
                ["a", "b"],
                name="fk2",
            )
            eq_(len(diffs), 2)
        elif hook_type == "name":
            eq_(
                {(d[0], d[1].name) for d in diffs},
                {("add_fk", "fk2"), ("add_fk", "fk1"), ("remove_fk", "fk2")},
            )


class AutogenerateFKOptionsTest(AutogenFixtureTest, TestBase):
    __backend__ = True

    def _fk_opts_fixture(self, old_opts, new_opts):
        m1 = MetaData()
        m2 = MetaData()

        Table(
            "some_table",
            m1,
            Column("id", Integer, primary_key=True),
            Column("test", String(10)),
        )

        Table(
            "user",
            m1,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("tid", Integer),
            ForeignKeyConstraint(["tid"], ["some_table.id"], **old_opts),
        )

        Table(
            "some_table",
            m2,
            Column("id", Integer, primary_key=True),
            Column("test", String(10)),
        )

        Table(
            "user",
            m2,
            Column("id", Integer, primary_key=True),
            Column("name", String(50), nullable=False),
            Column("tid", Integer),
            ForeignKeyConstraint(["tid"], ["some_table.id"], **new_opts),
        )

        return self._fixture(m1, m2)

    @config.requirements.fk_ondelete_is_reflected
    def test_add_ondelete(self):
        diffs = self._fk_opts_fixture({}, {"ondelete": "cascade"})

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            ondelete=None,
            conditional_name="servergenerated",
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            ondelete="cascade",
        )

    @config.requirements.fk_ondelete_is_reflected
    def test_remove_ondelete(self):
        diffs = self._fk_opts_fixture({"ondelete": "CASCADE"}, {})

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            ondelete="CASCADE",
            conditional_name="servergenerated",
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            ondelete=None,
        )

    def test_nochange_ondelete(self):
        """test case sensitivity"""
        diffs = self._fk_opts_fixture(
            {"ondelete": "caSCAde"}, {"ondelete": "CasCade"}
        )
        eq_(diffs, [])

    @config.requirements.fk_onupdate_is_reflected
    def test_add_onupdate(self):
        diffs = self._fk_opts_fixture({}, {"onupdate": "cascade"})

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            onupdate=None,
            conditional_name="servergenerated",
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            onupdate="cascade",
        )

    @config.requirements.fk_onupdate_is_reflected
    def test_remove_onupdate(self):
        diffs = self._fk_opts_fixture({"onupdate": "CASCADE"}, {})

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            onupdate="CASCADE",
            conditional_name="servergenerated",
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            onupdate=None,
        )

    @config.requirements.fk_onupdate
    def test_nochange_onupdate(self):
        """test case sensitivity"""
        diffs = self._fk_opts_fixture(
            {"onupdate": "caSCAde"}, {"onupdate": "CasCade"}
        )
        eq_(diffs, [])

    @config.requirements.fk_ondelete_restrict
    def test_nochange_ondelete_restrict(self):
        """test the RESTRICT option which MySQL doesn't report on"""

        diffs = self._fk_opts_fixture(
            {"ondelete": "restrict"}, {"ondelete": "restrict"}
        )
        eq_(diffs, [])

    @config.requirements.fk_onupdate_restrict
    def test_nochange_onupdate_restrict(self):
        """test the RESTRICT option which MySQL doesn't report on"""

        diffs = self._fk_opts_fixture(
            {"onupdate": "restrict"}, {"onupdate": "restrict"}
        )
        eq_(diffs, [])

    @config.requirements.fk_ondelete_noaction
    def test_nochange_ondelete_noaction(self):
        """test the NO ACTION option which generally comes back as None"""

        diffs = self._fk_opts_fixture(
            {"ondelete": "no action"}, {"ondelete": "no action"}
        )
        eq_(diffs, [])

    @config.requirements.fk_onupdate
    def test_nochange_onupdate_noaction(self):
        """test the NO ACTION option which generally comes back as None"""

        diffs = self._fk_opts_fixture(
            {"onupdate": "no action"}, {"onupdate": "no action"}
        )
        eq_(diffs, [])

    @config.requirements.fk_ondelete_restrict
    def test_change_ondelete_from_restrict(self):
        """test the RESTRICT option which MySQL doesn't report on"""

        # note that this is impossible to detect if we change
        # from RESTRICT to NO ACTION on MySQL.
        diffs = self._fk_opts_fixture(
            {"ondelete": "restrict"}, {"ondelete": "cascade"}
        )
        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            onupdate=None,
            ondelete=mock.ANY,  # MySQL reports None, PG reports RESTRICT
            conditional_name="servergenerated",
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            onupdate=None,
            ondelete="cascade",
        )

    @config.requirements.fk_ondelete_restrict
    def test_change_onupdate_from_restrict(self):
        """test the RESTRICT option which MySQL doesn't report on"""

        # note that this is impossible to detect if we change
        # from RESTRICT to NO ACTION on MySQL.
        diffs = self._fk_opts_fixture(
            {"onupdate": "restrict"}, {"onupdate": "cascade"}
        )
        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            onupdate=mock.ANY,  # MySQL reports None, PG reports RESTRICT
            ondelete=None,
            conditional_name="servergenerated",
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            onupdate="cascade",
            ondelete=None,
        )

    @config.requirements.fk_ondelete_is_reflected
    @config.requirements.fk_onupdate_is_reflected
    def test_ondelete_onupdate_combo(self):
        diffs = self._fk_opts_fixture(
            {"onupdate": "CASCADE", "ondelete": "SET NULL"},
            {"onupdate": "RESTRICT", "ondelete": "RESTRICT"},
        )

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            onupdate="CASCADE",
            ondelete="SET NULL",
            conditional_name="servergenerated",
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            onupdate="RESTRICT",
            ondelete="RESTRICT",
        )

    @config.requirements.fk_initially
    def test_add_initially_deferred(self):
        diffs = self._fk_opts_fixture({}, {"initially": "deferred"})

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            initially=None,
            conditional_name="servergenerated",
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            initially="deferred",
        )

    @config.requirements.fk_initially
    def test_remove_initially_deferred(self):
        diffs = self._fk_opts_fixture({"initially": "deferred"}, {})

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            initially="DEFERRED",
            deferrable=True,
            conditional_name="servergenerated",
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            initially=None,
        )

    @config.requirements.fk_deferrable
    @config.requirements.fk_initially
    def test_add_initially_immediate_plus_deferrable(self):
        diffs = self._fk_opts_fixture(
            {}, {"initially": "immediate", "deferrable": True}
        )

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            initially=None,
            conditional_name="servergenerated",
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            initially="immediate",
            deferrable=True,
        )

    @config.requirements.fk_deferrable
    @config.requirements.fk_initially
    def test_remove_initially_immediate_plus_deferrable(self):
        diffs = self._fk_opts_fixture(
            {"initially": "immediate", "deferrable": True}, {}
        )

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            initially=None,  # immediate is the default
            deferrable=True,
            conditional_name="servergenerated",
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            initially=None,
            deferrable=None,
        )

    @config.requirements.fk_initially
    @config.requirements.fk_deferrable
    def test_add_initially_deferrable_nochange_one(self):
        diffs = self._fk_opts_fixture(
            {"deferrable": True, "initially": "immediate"},
            {"deferrable": True, "initially": "immediate"},
        )

        eq_(diffs, [])

    @config.requirements.fk_initially
    @config.requirements.fk_deferrable
    def test_add_initially_deferrable_nochange_two(self):
        diffs = self._fk_opts_fixture(
            {"deferrable": True, "initially": "deferred"},
            {"deferrable": True, "initially": "deferred"},
        )

        eq_(diffs, [])

    @config.requirements.fk_initially
    @config.requirements.fk_deferrable
    def test_add_initially_deferrable_nochange_three(self):
        diffs = self._fk_opts_fixture(
            {"deferrable": None, "initially": "deferred"},
            {"deferrable": None, "initially": "deferred"},
        )

        eq_(diffs, [])

    @config.requirements.fk_deferrable
    def test_add_deferrable(self):
        diffs = self._fk_opts_fixture({}, {"deferrable": True})

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            deferrable=None,
            conditional_name="servergenerated",
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            deferrable=True,
        )

    @config.requirements.fk_deferrable_is_reflected
    def test_remove_deferrable(self):
        diffs = self._fk_opts_fixture({"deferrable": True}, {})

        self._assert_fk_diff(
            diffs[0],
            "remove_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            deferrable=True,
            conditional_name="servergenerated",
        )

        self._assert_fk_diff(
            diffs[1],
            "add_fk",
            "user",
            ["tid"],
            "some_table",
            ["id"],
            deferrable=None,
        )



import json
import os
import time
from typing import (
    Any,
    List,
    Optional,
)

from boss.model import User
from cfgtools import read
from message_handler_builder import Configs
from messagebus import (
    DataType,
    DefaultToken,
    Message,
    MessageBus,
    TokenCreator,
    TokenPriority,
)
from systools import sys_parsefilepath

TK_STORECFG_GET = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group="13",
    number="1",
)


class Configurator:

    def __init__(
        self,
        loader_path: str,
    ) -> None:

        self.loader_path = loader_path

        self.reader = read(file_or_path=loader_path)
        self.group = Configs.component_name.title() + "."
        self.configs = Configs()

    def load_configs(
        self,
    ) -> None:

        self.configs = Configs(
            otp_interval=self._get_otp_interval(),
            required_services=self._get_required_services(),
            api_retry_policy=self._get_api_retry_policy(),
            endpoints=self._get_endpoints(),
            database_path=self._get_databases_path(),
            default_users=self._get_default_users(),
        )

    def load_store_wide_configs(
        self,
        message_bus: MessageBus,
    ) -> None:

        self.configs.pump_credentials = Configs.PumpCredentials(
            base_url=self._read_sw_config(
                message_bus=message_bus,
                key="BackOffice.ServerURL",
            ),
            api_key=self._read_sw_config(
                message_bus=message_bus,
                key="BackOffice.ApiKey",
            ),
        )
        self.configs.store_id = self._read_sw_config(
            message_bus=message_bus,
            key="Store.Id",
        )

    def _get_endpoints(
        self,
    ) -> Configs.Endpoints:

        return Configs.Endpoints(
            authenticate=self._find(key="Endpoints.Authenticate"),
        )

    def _get_api_retry_policy(
        self,
    ) -> Configs.ApiRetryPolicy:

        return Configs.ApiRetryPolicy(
            count=int(
                self._find(
                    key="ApiRetryPolicy.Count",
                    default_value=1,
                )
            ),
            sleep=int(
                self._find(
                    key="ApiRetryPolicy.Sleep",
                    default_value=1,
                )
            ),
        )

    def _get_required_services(
        self,
    ) -> str:

        return "|".join(self.reader.find_values(path=self.group + "RequiredServices") or [])

    def _get_otp_interval(
        self,
    ) -> int:

        return int(
            self._find(
                key="OtpInterval",
                default_value=600,
            )
        )

    def _get_databases_path(
        self,
    ) -> str:

        database_path = self._find(
            key="DatabasePath",
            default_value="",
        )
        if database_path == "":
            comp_dir = os.path.dirname(self.loader_path)
            database_path = os.path.abspath(os.path.join(comp_dir, "..", "..", "databases", "users.db"))

        if database_path.startswith("file:") == -1:
            database_path = f"file:///{database_path}"

        return sys_parsefilepath(path=database_path)

    def _get_default_users(
        self,
    ) -> List[User]:

        default_users_file_path = self._find(
            key="DefaultUsersFilePath",
            default_value="",
        )

        if default_users_file_path == "":
            default_users_file_path = os.path.join(os.path.dirname(self.loader_path), "default_users.json")

        default_users_file_path = os.path.abspath(sys_parsefilepath(path=default_users_file_path))

        if not os.path.exists(default_users_file_path):
            return []

        default_users_dict = json.load(open(default_users_file_path))

        default_users: List[User] = []
        for user in default_users_dict:
            default_users.append(
                User(
                    user_id=user["UserId"],
                    user_name=user["UserName"],
                    long_name=user["LongName"],
                    password=user["Password"],
                    level=user["Level"],
                    status=user["Status"],
                    admission_date=user["AdmissionDate"],
                    termination_date=user["TerminationDate"],
                    fingerprint_minutiae=user["FingerPrintMinutiae"],
                    nfc_uuid=user["NfcUuid"],
                )
            )

        return default_users

    def _find(
        self,
        key: [str],
        default_value: Optional[Any] = None,
    ) -> Optional[str]:

        return self.reader.find_value(
            path=self.group + key,
            default=default_value,
        )

    @staticmethod
    def _read_sw_config(
        message_bus: MessageBus,
        key: str,
    ) -> str:

        for _ in range(0, 3):
            try:
                message = Message(
                    token=TK_STORECFG_GET,
                    data_type=DataType.param,
                    data=key.encode(),
                )
                msg = message_bus.send_message(
                    component="StoreWideConfig",
                    message=message,
                )
                if msg.token == DefaultToken.TK_SYS_ACK:
                    return msg.data.decode()
            except (Exception,):
                pass

            time.sleep(1)

        raise RuntimeError("Unable to get StoreWide configuration: {}".format(key))



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class Schemas(Enum):
    V1_23_1 = "https://opentelemetry.io/schemas/1.23.1"
    """
    The URL of the OpenTelemetry schema version 1.23.1.
    """

    V1_25_0 = "https://opentelemetry.io/schemas/1.25.0"
    """
    The URL of the OpenTelemetry schema version 1.25.0.
    """

    V1_26_0 = "https://opentelemetry.io/schemas/1.26.0"
    """
    The URL of the OpenTelemetry schema version 1.26.0.
    """

    V1_27_0 = "https://opentelemetry.io/schemas/1.27.0"
    """
    The URL of the OpenTelemetry schema version 1.27.0.
    """

    V1_28_0 = "https://opentelemetry.io/schemas/1.28.0"
    """
    The URL of the OpenTelemetry schema version 1.28.0.
    """

    V1_29_0 = "https://opentelemetry.io/schemas/1.29.0"
    """
    The URL of the OpenTelemetry schema version 1.29.0.
    """

    # when generating new semantic conventions,
    # make sure to add new versions version here.



import logging

from mw_helper import (
    show_message_dialog,
    show_filtered_list_box_dialog
)
from sysactions import (
    show_messagebox,
    StopAction
)
from typing import (
    List,
    Optional
)

from application.interactor import OrderInteractor

logger = logging.getLogger("PosActions")


def display_coupons_from_report(
    pos_id,         # type: str
    initial_date,   # type: str
    end_date,       # type: str
    operator_id     # type: Optional[str]
):
    # type: (...) -> List

    try:
        if initial_date is None or end_date is None:
            raise StopAction()

        benefits = OrderInteractor(pos_id).get_orders_benefit(initial_date, end_date, operator_id)
        if benefits is None or len(benefits) == 1:
            show_messagebox(pos_id, "$ORDER_HAS_NO_APPLIED_DISCOUNT")
            raise StopAction()

        applied_benefits = [benefit["name"] for benefit in benefits]
        sorted_applied_benefits = [applied_benefits[0]] + sorted(list(dict.fromkeys(applied_benefits[1:])))
        options = ["{0}##True".format(benefit) for benefit in sorted_applied_benefits]
        selected_index = show_filtered_list_box_dialog(pos_id, options, "$BENEFIT_REPORT", mask="NOFILTER")

        if selected_index is None:
            raise StopAction()

        chosen_benefit = [x for x in benefits if x["name"] == sorted_applied_benefits[selected_index]]
        return chosen_benefit
    except StopAction:
        raise
    except Exception as ex:
        show_message_dialog(pos_id, "$ERROR", ex.message)
        logger.exception("Error display coupons from report")
        raise StopAction()



import os

import cfgtools

change_number_seats_sequence = None


def can_change_number_seats_sequence():
    # type: (...) -> bool

    global change_number_seats_sequence

    if change_number_seats_sequence is None:
        config = cfgtools.read(os.environ["LOADERCFG"])
        change_number_seats_sequence = config.find_value(
            path="Customizations.ChangeNumberSeatsSequence",
            default="false",
        )

    return change_number_seats_sequence.lower() == "true"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: opentelemetry/proto/resource/v1/resource.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from opentelemetry.proto.common.v1 import (
    common_pb2 as opentelemetry_dot_proto_dot_common_dot_v1_dot_common__pb2,
)


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n.opentelemetry/proto/resource/v1/resource.proto\x12\x1fopentelemetry.proto.resource.v1\x1a*opentelemetry/proto/common/v1/common.proto"i\n\x08Resource\x12;\n\nattributes\x18\x01 \x03(\x0b\x32\'.opentelemetry.proto.common.v1.KeyValue\x12 \n\x18\x64ropped_attributes_count\x18\x02 \x01(\rB\x83\x01\n"io.opentelemetry.proto.resource.v1B\rResourceProtoP\x01Z*go.opentelemetry.io/proto/otlp/resource/v1\xaa\x02\x1fOpenTelemetry.Proto.Resource.V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "opentelemetry.proto.resource.v1.resource_pb2", _globals
)
if not _descriptor._USE_C_DESCRIPTORS:
    _globals["DESCRIPTOR"]._loaded_options = None
    _globals["DESCRIPTOR"]._serialized_options = (
        b'\n"io.opentelemetry.proto.resource.v1B\rResourceProtoP\001Z*go.opentelemetry.io/proto/otlp/resource/v1\252\002\037OpenTelemetry.Proto.Resource.V1'
    )
    _globals["_RESOURCE"]._serialized_start = 127
    _globals["_RESOURCE"]._serialized_end = 232
# @@protoc_insertion_point(module_scope)



from logging import Logger

from mw_helper import (
    ensure_list,
    convert_to_dict,
)

from production.repository import ProductRepository
from typing import List, Optional, Union

from ._OrderChangerProductionBox import OrderChangerProductionBox
from ._ProductionBox import ProductionBox
from ..model import (
    ProductionOrder,
    Item,
)


class NoJitAddTagBox(OrderChangerProductionBox):

    def __init__(self, name, sons, product_repository, tag, logger=None):
        # type: (str, Optional[Union[List[ProductionBox], ProductionBox]], ProductRepository, List[str], Logger) -> None

        self.part_codes = convert_to_dict(ensure_list(product_repository.get_no_jit_part_codes()))
        self.tag = tag
        super(NoJitAddTagBox, self).__init__(name, sons, logger)

    def change_order(
        self,   # type: NoJitAddTagBox
        order,  # type: ProductionOrder
    ):
        # type: (...) -> ProductionOrder

        for item in order.items:
            self.add_tag_on_item(item=item)

        return order

    def add_tag_on_item(
        self,   # type: NoJitAddTagBox
        item,   # type: Item
    ):
        # type: (...) -> None

        all_sons_tagged = True
        for son in item.items:
            if not self.add_tag_on_item(item=son):
                all_sons_tagged = False

        item_has_no_jit = item.part_code in self.part_codes and item.does_not_have_tag(tag=self.tag)
        if item_has_no_jit and (not item.is_combo() or all_sons_tagged):
            item.add_tag(tag=self.tag)



# encoding: utf-8

import json
from xml.etree import cElementTree as eTree

from msgbus import (
    FM_PARAM,
    TK_CMP_MESSAGE,
    TK_SLCTRL_OMGR_ORDERPICT,
    TK_SYS_NAK,
)

import requests
from application.model import (
    ApplierSubTypes,
    TotaledMessage,
    TranslatedMessage,
)
from application.model.configuration import Configurations
from application.model.customexception import (
    NotFound,
    ForbiddenError,
    ProcessorError,
    InUseError,
    NotAvailableError,
    NotFoundError,
    NoResponseError,
    AlreadyUnlockedError,
    NotAllowedError,
    ClientCustomError,
    ResponseError,
    GoneError,
    ServiceNotFoundError,
)
from application.model.orderinfo import (
    Order,
    OrderJsonEncoder,
)
from helper import retry
from messagebus import (
    Message,
    MessageBus,
)
from requests import Response
from typing import (
    Any,
    Optional,
    List,
    Tuple,
    Dict,
    Union,
)
from util import get_applier_sub_type_from_order
from posentitiesformatter import (
    OrderFormatter,
    TableFormatter,
)


class APIRepository(object):

    def __init__(
        self,
        configs,          # type: Configurations
        message_bus,      # type: MessageBus
    ):
        # type: (...) -> None

        self.logger = configs.logger
        self.configs = configs
        self.message_bus = message_bus
        self.exception_map = {
            204: NoResponseError,
            206: NoResponseError,
            403: ForbiddenError,
            404: NotFoundError,
            405: NotAllowedError,
            410: GoneError,
            500: ProcessorError,
        }
        self.per_method_exception_map = {
            "lock": {
                404: NotFoundError,
                406: ClientCustomError,
                409: InUseError,
            },
            "burn": {
                400: InUseError,
                406: NotAvailableError,
                409: AlreadyUnlockedError
            }
        }

    def retrieve_loyalty_customer_info(self, loyalty_customer_id):
        # type: (str) -> Optional[str, Exception]
        try:
            url = self._get_customer_url(loyalty_customer_id)
            headers = self._get_headers(applier_sub_type=ApplierSubTypes.LOYALTY)
            body = dict(storeId=self.configs.store_id)
            errors_list = []

            response = retry(
                logger=self.logger,
                method_name="retrieve_loyalty_customer_info",
                n=self.configs.partner_retry_policy.count,
                sleep_seconds=self.configs.partner_retry_policy.sleep,
                trying_to_do=lambda: requests.request(
                    method="GET",
                    url=url,
                    json=body,
                    headers=headers,
                    timeout=5,
                ),
                is_success=lambda resp: resp.status_code == 200,
                success=lambda resp: resp,
                failed=lambda resp: errors_list.append((resp.status_code, resp)),
                on_exception=lambda ex: errors_list.append(ex)
            )
            if response is not None:
                return response.content

            response_error = self._get_response_error(errors_list)
            raise self._build_lock_custom_exception(
                status_code=response_error.status_code,
                response_error=response_error,
            )
        except Exception as ex:
            self.logger.error("Error on get_benefit: {}".format(ex.message.encode("utf-8")))
            raise

    def retrieve_coupon_benefit(
        self,       # type: APIRepository
        pos_id,     # type: int
        coupon_id,  # type: str
    ):
        # type: (...) -> Optional[str]

        order_picture = self.get_order_picture(
            pos_id=pos_id,
        )
        table_info = self._get_table_info_associated_with_order_picture(
            pos_id=pos_id,
            order_picture=order_picture,
        )
        order_info = self._build_order_info(
            order_picture=order_picture,
            table_picture=table_info,
        )

        self.logger.info("Order json sent on retrieve:\n{}".format(json.dumps(order_info)))
        benefit_raw = self._get_coupon_benefit(
            benefit_id=coupon_id,
            order=order_info,
        )

        if not benefit_raw:
            self.logger.info("No benefit found")
            return None

        benefit = json.loads(benefit_raw)
        self.logger.info("Json from partner: {}".format(benefit_raw))
        benefit = benefit.get("benefit", benefit)

        notify = benefit.get("notify", False)
        benefit_id = benefit.get("id", None)
        if notify and benefit_id:
            lock_id = self.lock_benefit(
                benefit=benefit,
                pos_id=pos_id,
                applier_sub_type=ApplierSubTypes.COUPON,
            )
            benefit["lockId"] = lock_id

        benefit["applier"] = "Loyalty"
        benefit["applierSubType"] = ApplierSubTypes.COUPON
        return json.dumps(benefit)

    def _get_table_info(
        self,       # type: APIRepository
        pos_id,     # type: int
        table_id,   # type: int
    ):
        # type: (...) -> Optional[str]

        params = ("TABLE_INFO", pos_id, table_id)
        req_buf = "\0".join([str(p) for p in params])
        message = Message(
            token=TK_CMP_MESSAGE,
            data_type=FM_PARAM,
            data=req_buf,
        )
        msg = self.message_bus.send_message(
            component='TABLEMGR',
            message=message,
        )

        if msg.token != TK_SYS_NAK:
            return msg.data
        return None

    def get_order_picture(
        self,               # type: APIRepository
        pos_id,             # type: Union[str, int]
        order_id="",        # type: Union[str, int]
        no_cache=None,      # type: Optional[str]
    ):
        # type: (...) -> Optional[eTree]

        params = (pos_id, order_id, 'true' if no_cache else 'false')
        req_buf = '\x00'.join((str(p) for p in params))

        message = Message(
            token=TK_SLCTRL_OMGR_ORDERPICT,
            data_type=FM_PARAM,
            data=req_buf,
        )
        msg = self.message_bus.send_message(
            component='ORDERMGR' + str(pos_id),
            message=message,
        )

        if msg.token == TK_SYS_NAK:
            return None

        return msg.data.split('\0')[2]

    def _get_table_info_associated_with_order_picture(
        self,           # type: APIRepository
        pos_id,         # type: int
        order_picture,  # type: str
    ):
        table_info = None
        order_info = OrderFormatter().get_order_picture_dict(order_picture=eTree.fromstring(order_picture))
        custom_properties = order_info.get('CustomOrderProperties')
        table_id = custom_properties.get('TABLE_ID') if custom_properties and "TABLE_ID" in custom_properties else None
        if table_id is not None:
            table_info = self._get_table_info(
                pos_id=pos_id,
                table_id=table_id,
            )
        return table_info

    def _build_order_info(
        self,                   # type: APIRepository
        order_picture,          # type: str
        table_picture=None,     # type: Optional[str]
        payment_info=None,      # type: Optional[Dict]
    ):
        # type: (...) -> dict

        order_info = OrderFormatter().get_order_picture_dict(order_picture=eTree.fromstring(order_picture))
        self.fix_order_info_to_include_api_required_fields(order_info)

        table_info = None
        if table_picture:
            table_info = TableFormatter().get_table_picture_dict(table_picture=eTree.fromstring(table_picture))

        order = Order(
            store_id=self.configs.store_id,
            order_info=order_info,
            table_info=table_info,
            payment_info=payment_info,
        )
        return json.loads(json.dumps(order, encoding="utf-8", cls=OrderJsonEncoder))

    @staticmethod
    def fix_order_info_to_include_api_required_fields(order_info):
        order_info["instance"] = 0
        order_info["priceBasis"] = "G"

    def burn_benefit(
        self,                   # type: APIRepository
        benefit_id,             # type: str
        order_picture,          # type: str
        table_picture=None,     # type: Optional[str]
    ):
        # type: (...) -> None

        applier_sub_type = get_applier_sub_type_from_order(
            benefit_id=benefit_id,
            order_picture=order_picture,
            logger=self.logger,
        )

        url = self._get_burn_url(
            benefit_id=benefit_id,
            applier_sub_type=applier_sub_type,
        )
        headers = self._get_headers(applier_sub_type=applier_sub_type)
        errors_list = []

        body_json = self._build_order_info(
            order_picture=order_picture,
            table_picture=table_picture,
        )
        self.logger.debug("Order json sent on burn:\n{}".format(body_json))

        response = retry(
            logger=self.logger,
            method_name="burn_benefit",
            n=self.configs.partner_retry_policy.count,
            sleep_seconds=self.configs.partner_retry_policy.sleep,
            trying_to_do=lambda: requests.request(
                method="PUT",
                url=url,
                headers=headers,
                json=body_json,
                timeout=20,
            ),
            is_success=lambda resp: resp.status_code in [200, 204],
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp))
        )

        # TODO: Treat burnId in case of 200
        if response is not None:
            return

        if not errors_list:
            raise NotFound()

        response_error = self._get_response_error(errors_list)
        raise self._build_unlock_burn_custom_exception(
            status_code=response_error.status_code,
            response_error=response_error
        )

    def _get_coupon_benefit(
        self,               # type: APIRepository
        benefit_id,         # type: str
        order,              # type: Dict
    ):
        # type: (...) -> str

        url = self._get_benefit_url(
            voucher_id=benefit_id,
            applier_sub_type=ApplierSubTypes.COUPON,
        )
        headers = self._get_headers(applier_sub_type=ApplierSubTypes.COUPON)
        errors_list = []

        response = retry(
            logger=self.logger,
            method_name="get_benefit",
            n=1,
            sleep_seconds=1,
            trying_to_do=lambda: requests.request(
                "POST",
                url=url,
                headers=headers,
                timeout=20,
                json=order,
            ),
            is_success=lambda resp: resp.status_code == 200,
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp))
        )

        if response:
            return response.content
        else:
            if not errors_list:
                raise NotFound()

            response_error = self._get_response_error(errors_list)
            raise self._build_lock_custom_exception(
                status_code=response_error.status_code,
                response_error=response_error,
            )

    def lock_benefit(
        self,                   # type: APIRepository
        benefit,                # type: Dict[str, Any]
        pos_id,                 # type: int
        applier_sub_type=None,  # type: Optional[str]
        payment_info=None,      # type: Optional[Dict[str, Any]]
    ):
        # type: (...) -> str

        benefit_id = benefit.get("id")
        applier_sub_type = applier_sub_type or benefit.get("applierSubType")

        url = self._get_lock_url(
            benefit_id=benefit_id,
            applier_sub_type=applier_sub_type,
        )
        headers = self._get_headers(applier_sub_type=applier_sub_type)
        errors_list = []

        order_picture = self.get_order_picture(pos_id=pos_id)
        table_info = self._get_table_info_associated_with_order_picture(
            pos_id=pos_id,
            order_picture=order_picture,
        )

        body_json = self._build_order_info(
            order_picture=order_picture,
            table_picture=table_info,
            payment_info=payment_info,
        )

        response = retry(
            logger=self.logger,
            method_name="lock_benefit",
            n=1,
            sleep_seconds=1,
            trying_to_do=lambda: requests.request("PUT", url=url, headers=headers, json=body_json, timeout=20),
            is_success=lambda resp: resp.status_code in [200, 204],
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp))
        )

        if response:
            return self._get_lock_id_from_response(
                response=response,
            )

        if not errors_list:
            raise NotFound()

        response_error = self._get_response_error(errors_list=errors_list)
        raise self._build_lock_custom_exception(
            status_code=response_error.status_code,
            response_error=response_error,
        )

    def unlock_benefit(
        self,           # type: APIRepository
        benefit_id,     # type: str
        benefit,        # type: Dict[str, Any]
        order_picture,  # type: str
        table_picture,  # type: str
    ):
        # type: (...) -> None

        if benefit:
            applier_sub_type = benefit.get("applierSubType")
        else:
            applier_sub_type = get_applier_sub_type_from_order(
                benefit_id=benefit_id,
                order_picture=order_picture,
                logger=self.logger,
            )
        url = self._get_unlock_url(
            benefit_id=benefit_id,
            applier_sub_type=applier_sub_type,
        )
        headers = self._get_headers(applier_sub_type=applier_sub_type)
        errors_list = []

        body_json = self._build_order_info(
            order_picture=order_picture,
            table_picture=table_picture,
        )

        response = retry(
            logger=self.logger,
            method_name="unlock_benefit",
            n=self.configs.partner_retry_policy.count,
            sleep_seconds=self.configs.partner_retry_policy.sleep,
            trying_to_do=lambda: requests.request(
                method="PUT",
                url=url,
                json=body_json,
                headers=headers,
                timeout=20,
            ),
            is_success=lambda resp: resp.status_code in [200, 204],
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp))
        )

        if response is not None:
            return

        if not errors_list:
            raise NotFound()

        response_error = self._get_response_error(errors_list)
        raise self._build_unlock_burn_custom_exception(
            status_code=response_error.status_code,
            response_error=response_error
        )

    def get_totaled_message(
        self,           # type: APIRepository
        pos_id,         # type: int
        customer_id,    # type: str
    ):
        # type: (...) -> Optional[TotaledMessage]

        try:
            url = self._get_totaled_url(customer_id)
            errors_list = []

            order_picture = self.get_order_picture(
                pos_id=pos_id,
            )
            table_info = self._get_table_info_associated_with_order_picture(
                pos_id=pos_id,
                order_picture=order_picture,
            )

            body_json = self._build_order_info(
                order_picture=order_picture,
                table_picture=table_info,
            )

            self.logger.debug("Order json sent on get_totaled_message:\n{}".format(body_json))

            response = retry(
                logger=self.logger,
                method_name="get_totaled_message",
                n=self.configs.partner_retry_policy.count,
                sleep_seconds=self.configs.partner_retry_policy.sleep,
                trying_to_do=lambda: requests.request(
                    method="POST",
                    url=url,
                    json=body_json,
                    headers=self._get_headers(applier_sub_type=ApplierSubTypes.LOYALTY),
                    timeout=5,
                ),
                is_success=lambda resp: resp.status_code in (200, 404),
                success=lambda resp: resp,
                failed=lambda resp: errors_list.append((resp.status_code, resp)),
                on_exception=lambda ex: errors_list.append(ex)
            )
            if response is not None:
                if response.status_code == 404:
                    return None
                response_dict = json.loads(response.content)
                translated_message = self._build_translated_message(
                    body_dict=response_dict,
                    message_field="totalizationMessage"
                )
                return TotaledMessage(totalization_message=translated_message)

            response_error = self._get_response_error(errors_list=errors_list)
            raise ResponseError(
                logger=self.logger,
                response=response_error,
            )
        except Exception as ex:
            self.logger.error("Error on get_totaled_message: {}".format(ex))
            raise

    def _get_totaled_url(
        self,           # type: APIRepository
        customer_id,    # type: str
    ):
        # type: (...) -> str

        end_point_url = self.configs.endpoints.get_totaled_message.format(customerId=customer_id)
        full_url = self.configs.credentials[ApplierSubTypes.LOYALTY].base_url + end_point_url

        self.logger.info("Requesting from totaled endpoint: {}".format(full_url))

        return full_url

    def get_benefit_on_payment(
        self,                   # type: APIRepository
        pos_id,                 # type: int
        loyalty_customer_id,    # type: str
        payment_info,           # type: Dict[str, Union[str, int, float]]
    ):
        # type: (...) -> Optional[Dict]

        url = self._get_payment_url(loyalty_customer_id=loyalty_customer_id)
        if not url:
            raise ServiceNotFoundError()

        headers = self._get_headers(applier_sub_type=ApplierSubTypes.PAYMENT)
        errors_list = []

        order_picture = self.get_order_picture(
            pos_id=pos_id,
        )
        table_info = self._get_table_info_associated_with_order_picture(
            pos_id=pos_id,
            order_picture=order_picture,
        )

        body_json = self._build_order_info(
            order_picture=order_picture,
            table_picture=table_info,
            payment_info=payment_info,
        )

        response = retry(
            logger=self.logger,
            method_name="payment",
            n=self.configs.partner_retry_policy.count,
            sleep_seconds=self.configs.partner_retry_policy.sleep,
            trying_to_do=lambda: requests.request(
                method="POST",
                url=url,
                json=body_json,
                headers=headers,
                timeout=20,
            ),
            is_success=lambda resp: resp.status_code in (200, 404),
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp)),
            on_exception=lambda ex: errors_list.append(ex),
        )

        if response is not None:
            if response.status_code == 404:
                return None
            return json.loads(response.content)

        response_error = self._get_response_error(errors_list=errors_list)
        raise ResponseError(
            logger=self.logger,
            response=response_error,
        )

    def end_of_sale_notes(
        self,           # type: APIRepository
        pos_id,         # type: int
        order_id,       # type: int
    ):
        # type: (...) -> Optional[List[Dict[str, str]]]

        url = self._get_end_of_sale_notes_url()
        if not url:
            return None

        headers = self._get_headers(applier_sub_type=ApplierSubTypes.MARKETING)
        errors_list = []

        order_picture = self.get_order_picture(
            pos_id=pos_id,
            order_id=order_id,

        )
        body_json = self._build_order_info(order_picture=order_picture)
        self.logger.info("Order json sent on end of sale:\n{}".format(json.dumps(body_json)))

        response = retry(
            logger=self.logger,
            method_name="end_of_sale_notes",
            n=1,
            sleep_seconds=self.configs.partner_retry_policy.sleep,
            trying_to_do=lambda: requests.request(
                method="POST",
                url=url,
                json=body_json,
                headers=headers,
                timeout=self.configs.credentials[ApplierSubTypes.MARKETING].timeout,
            ),
            is_success=lambda resp: resp.status_code in [200, 204, 404],
            success=lambda resp: resp,
            failed=lambda resp: errors_list.append((resp.status_code, resp)),
            on_exception=lambda ex: errors_list.append(ex),
        )

        if response is not None:
            if response.status_code in [204, 404]:
                return None
            return json.loads(response.content)

        response_error = self._get_response_error(errors_list=errors_list)
        raise ResponseError(
            logger=self.logger,
            response=response_error,
        )

    def _get_customer_url(self, loyalty_customer_id):
        # type: (str) -> str

        end_point_url = self.configs.endpoints.retrieve_loyalty_customer_info.format(customerId=loyalty_customer_id)
        full_url = self.configs.credentials[ApplierSubTypes.LOYALTY].base_url + end_point_url

        self.logger.info("Requesting from get customer info endpoint: {}".format(full_url))

        return full_url

    def _get_benefit_url(
        self,               # type: APIRepository
        voucher_id,         # type: str
        applier_sub_type,   # type: str
    ):
        # type: (...) -> str

        end_point_url = self.configs.endpoints.get_benefit.format(voucherId=voucher_id)
        full_url = self.configs.credentials[applier_sub_type].base_url + end_point_url

        self.logger.info("Requesting from get benefit endpoint: {}".format(full_url))

        return full_url

    def _get_end_of_sale_notes_url(
        self,   # type: APIRepository
    ):
        # type: (...) -> Optional[str]

        base_url = self.configs.credentials[ApplierSubTypes.MARKETING].base_url
        if not base_url:
            return None

        end_point_url = self.configs.endpoints.end_of_sale_notes
        full_url = base_url + end_point_url

        return full_url

    def _get_burn_url(
        self,               # type: APIRepository
        benefit_id,         # type: str
        applier_sub_type,   # type: str
    ):
        # type: (...) -> str

        end_point_url = self.configs.endpoints.burn_benefit.format(benefitId=benefit_id)
        full_url = self.configs.credentials[applier_sub_type].base_url + end_point_url

        self.logger.info("Requesting from burn endpoint: {}".format(full_url))

        return full_url

    def _get_lock_url(
        self,               # type: APIRepository
        benefit_id,         # type: str
        applier_sub_type,   # type: str
    ):
        # type: (...) -> str

        end_point_url = self.configs.endpoints.lock_benefit.format(benefitId=benefit_id)
        full_url = self.configs.credentials[applier_sub_type].base_url + end_point_url

        self.logger.info("Requesting from lock endpoint: {}".format(full_url))

        return full_url

    def _get_unlock_url(
        self,               # type: APIRepository
        benefit_id,         # type: str
        applier_sub_type,   # type: str
    ):
        # type: (...) -> str

        end_point_url = self.configs.endpoints.unlock_benefit.format(benefitId=benefit_id)
        full_url = self.configs.credentials[applier_sub_type].base_url + end_point_url

        self.logger.info("Requesting from unlock endpoint: {}".format(full_url))

        return full_url

    def _get_payment_url(
        self,                   # type: APIRepository
        loyalty_customer_id,    # type: str
    ):
        # type: (...) -> str

        base_url = self.configs.credentials[ApplierSubTypes.PAYMENT].base_url
        if not base_url:
            return ""

        end_point_url = self.configs.endpoints.payment.format(customerId=loyalty_customer_id)
        full_url = base_url + end_point_url

        self.logger.info("Requesting from payment endpoint: {}".format(full_url))

        return full_url

    def _get_headers(
        self,               # type: APIRepository
        applier_sub_type,   # type: str
    ):
        # type: (...) -> Dict

        return {"X-Api-Key": self.configs.credentials[applier_sub_type].api_key, "Content-type": "application/json"}

    def _build_per_method_custom_exception(
        self,            # type: APIRepository
        method_name,     # type: str
        status_code,     # type: int
        response_error,  # type: Response
    ):
        # type: (...) -> Optional[Exception]

        exception_cls = self.exception_map.get(
            status_code,
            self.per_method_exception_map.get(method_name, ResponseError).get(status_code, ResponseError)
        )
        if exception_cls == ClientCustomError:
            return self._build_client_custom_error(response_error=response_error)

        return exception_cls(
            logger=self.logger,
            response=response_error,
        )

    def _build_client_custom_error(
        self,            # type: APIRepository
        response_error,  # type: Response
    ):
        body = response_error.json()
        return ClientCustomError(
            message=self._build_translated_message(
                body_dict=body,
                message_field="message",
            )
        )

    def _build_translated_message(
        self,               # type: APIRepository
        body_dict,          # type: Dict
        message_field,      # type: str
    ):
        # type: (...) -> TranslatedMessage

        try:
            return TranslatedMessage(
                pt=body_dict[message_field].get("pt", ""),
                en=body_dict[message_field].get("en", ""),
                es=body_dict[message_field].get("es", ""),
            )
        except Exception as ex:
            self.logger.exception(msg="Invalid response body from partner - {}".format(body_dict), exc_info=ex)
            return TranslatedMessage(
                pt="",
                en="",
                es="",
            )

    def _build_lock_custom_exception(
        self,            # type: APIRepository
        status_code,     # type: int
        response_error,  # type: Response
    ):
        # type: (...) -> Exception

        return self._build_per_method_custom_exception(
            method_name="lock",
            status_code=status_code,
            response_error=response_error,
        )

    def _build_unlock_burn_custom_exception(
        self,            # type: APIRepository
        status_code,     # type: int
        response_error,  # type: Response
    ):
        # type: (...) -> Exception

        return self._build_per_method_custom_exception(
            method_name="burn",
            status_code=status_code,
            response_error=response_error,
        )

    @staticmethod
    def _get_response_error(
        errors_list,    # type: List[Tuple[int, Response]]
    ):
        # type: (...) -> Optional[Response]

        if not errors_list:
            return None

        for last_error in reversed(errors_list):
            if not isinstance(last_error, Exception):
                _, response = last_error
                return response

        raise errors_list[-1]

    @staticmethod
    def _get_lock_id_from_response(
        response,   # type: Response
    ):
        # type: (...) -> str

        if response.status_code == 200:
            content = json.loads(response.content or "{}")
            return content.get("lockId")

        return ""



# -*- coding: utf-8 -*-
import base64
import logging
import time
from xml.etree import cElementTree as eTree

from old_helper import OrderTaker
from pos_model import OrderKey
from posot import OrderTakerException

from nfcedisabler import (
    NfceDisabler,
    OrderRepository,
)
from requests.exceptions import RequestException
from services import OrderService

logger = logging.getLogger("FiscalWrapperNfceDisabler")


class OrderDisabler(object):

    def __init__(
        self,
        order_repository,   # type: OrderRepository
        nfce_disabler,      # type: NfceDisabler
        order_taker,        # type: OrderTaker
    ):
        # type: (...) -> None

        self.order_repository = order_repository
        self.nfce_disabler = nfce_disabler
        self.order_taker = order_taker

        self.order_service = OrderService(mb_context=order_repository.mbcontext)

    def disable_all_orders(
        self,   # type: OrderDisabler
    ):
        # type: (...) -> None

        logger.info("Inutilizando notas")
        order_keys = self.order_repository.get_all_orders_to_disable()
        logger.info("Orders para inutilizar: {0}".format(len(order_keys)))

        for order_key in order_keys:  # type: OrderKey
            pos_id = order_key.pos_id
            try:
                logger.info("Inutilizando posId/orderId: {}/{}".format(pos_id, order_key.order_id))
                order_xml = self.order_taker.get_order_picture(pos_id, order_key.order_id)

                logger.info("Iniciando inutilização da order {0} do pos {1}".format(order_key.order_id, pos_id))
                xml = self.nfce_disabler.disable_fiscal_number(eTree.XML(order_xml))
                custom_properties = {
                    "ORDER_DISABLED": "true",
                    "DISABLED_FISCAL_XML": base64.b64encode(xml),
                }
                self.order_service.set_custom_properties(
                    pos_id=pos_id,
                    order_id=order_key.order_id,
                    custom_properties=custom_properties,
                )
                logger.info("Order inutilizada com sucesso: {0}".format(order_key.order_id))
            except RequestException:
                logger.exception("Erro de Conexao com a Sefaz tentando inutulizar nota, tentaremos novamente...")
            except OrderTakerException:
                logger.exception("Order Taker Error, tentaremos novamente...")
            except (Exception,):
                message = "Erro inutilizando nota, verifique o Layout e os parametros enviados pela Sefaz, "\
                    "não tentaremos novamente."
                logger.exception(message)
                self.order_service.set_custom_property(
                    pos_id=pos_id,
                    order_id=order_key.order_id,
                    key="ORDER_DISABLED",
                    value="false",
                )
            finally:
                time.sleep(5)



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.  All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file or at
# https://developers.google.com/open-source/licenses/bsd

"""Encoding related utilities."""
import re

def _AsciiIsPrint(i):
  return i >= 32 and i < 127

def _MakeStrEscapes():
  ret = {}
  for i in range(0, 128):
    if not _AsciiIsPrint(i):
      ret[i] = r'\%03o' % i
  ret[ord('\t')] = r'\t'  # optional escape
  ret[ord('\n')] = r'\n'  # optional escape
  ret[ord('\r')] = r'\r'  # optional escape
  ret[ord('"')] = r'\"'  # necessary escape
  ret[ord('\'')] = r"\'"  # optional escape
  ret[ord('\\')] = r'\\'  # necessary escape
  return ret

# Maps int -> char, performing string escapes.
_str_escapes = _MakeStrEscapes()

# Maps int -> char, performing byte escaping and string escapes
_byte_escapes = {i: chr(i) for i in range(0, 256)}
_byte_escapes.update(_str_escapes)
_byte_escapes.update({i: r'\%03o' % i for i in range(128, 256)})


def _DecodeUtf8EscapeErrors(text_bytes):
  ret = ''
  while text_bytes:
    try:
      ret += text_bytes.decode('utf-8').translate(_str_escapes)
      text_bytes = ''
    except UnicodeDecodeError as e:
      ret += text_bytes[:e.start].decode('utf-8').translate(_str_escapes)
      ret += _byte_escapes[text_bytes[e.start]]
      text_bytes = text_bytes[e.start+1:]
  return ret


def CEscape(text, as_utf8) -> str:
  """Escape a bytes string for use in an text protocol buffer.

  Args:
    text: A byte string to be escaped.
    as_utf8: Specifies if result may contain non-ASCII characters.
        In Python 3 this allows unescaped non-ASCII Unicode characters.
        In Python 2 the return value will be valid UTF-8 rather than only ASCII.
  Returns:
    Escaped string (str).
  """
  # Python's text.encode() 'string_escape' or 'unicode_escape' codecs do not
  # satisfy our needs; they encodes unprintable characters using two-digit hex
  # escapes whereas our C++ unescaping function allows hex escapes to be any
  # length.  So, "\0011".encode('string_escape') ends up being "\\x011", which
  # will be decoded in C++ as a single-character string with char code 0x11.
  text_is_unicode = isinstance(text, str)
  if as_utf8:
    if text_is_unicode:
      return text.translate(_str_escapes)
    else:
      return _DecodeUtf8EscapeErrors(text)
  else:
    if text_is_unicode:
      text = text.encode('utf-8')
    return ''.join([_byte_escapes[c] for c in text])


_CUNESCAPE_HEX = re.compile(r'(\\+)x([0-9a-fA-F])(?![0-9a-fA-F])')


def CUnescape(text: str) -> bytes:
  """Unescape a text string with C-style escape sequences to UTF-8 bytes.

  Args:
    text: The data to parse in a str.
  Returns:
    A byte string.
  """

  def ReplaceHex(m):
    # Only replace the match if the number of leading back slashes is odd. i.e.
    # the slash itself is not escaped.
    if len(m.group(1)) & 1:
      return m.group(1) + 'x0' + m.group(2)
    return m.group(0)

  # This is required because the 'string_escape' encoding doesn't
  # allow single-digit hex escapes (like '\xf').
  result = _CUNESCAPE_HEX.sub(ReplaceHex, text)

  # Replaces Unicode escape sequences with their character equivalents.
  result = result.encode('raw_unicode_escape').decode('raw_unicode_escape')
  # Encode Unicode characters as UTF-8, then decode to Latin-1 escaping
  # unprintable characters.
  result = result.encode('utf-8').decode('unicode_escape')
  # Convert Latin-1 text back to a byte string (latin-1 codec also works here).
  return result.encode('latin-1')



# cui.py - functions for handling Peruvian personal numbers
# coding: utf-8
#
# Copyright (C) 2019 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""CUI (Cédula Única de Identidad, Peruvian identity number).

The Cédula Única de Identidad (CUI) is the unique identifier for persons that
appears on the Documento Nacional de Identidad (DNI), the national identity
document of Peru. The number consists of 8 digits and an optional extra check
digit.

More information:

* https://www.gob.pe/235-documento-nacional-de-identidad-dni
* https://es.wikipedia.org/wiki/Documento_Nacional_de_Identidad_(Perú)

>>> validate('10117410')
'10117410'
>>> validate('10117410-2')
'101174102'
>>> validate('10117410-3')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> to_ruc('10117410-2')
'10101174102'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').strip().upper()


def calc_check_digits(number):
    """Calculate the possible check digits for the CUI."""
    number = compact(number)
    weights = (3, 2, 7, 6, 5, 4, 3, 2)
    c = sum(w * int(n) for w, n in zip(weights, number)) % 11
    return '65432110987'[c] + 'KJIHGFEDCBA'[c]


def to_ruc(number):
    """Convert the number to a valid RUC."""
    from stdnum.pe import ruc
    number = '10' + compact(number)[:8]
    return number + ruc.calc_check_digit(number)


def validate(number):
    """Check if the number provided is a valid CUI. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if len(number) not in (8, 9):
        raise InvalidLength()
    if not isdigits(number[:8]):
        raise InvalidFormat()
    if len(number) > 8 and number[-1] not in calc_check_digits(number):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is a valid CUI. This checks the length,
    formatting and check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from typing import (
    Optional,
    List,
)


class ServiceSeats(object):

    def __init__(
        self,
        seats,                 # type: str
        custom_seats=None,     # type: Optional[List[str]]
    ):
        # type: (...) -> None

        self.seats = seats
        self.custom_seats = custom_seats



# nit.py - functions for handling El Salvador NIT numbers
# coding: utf-8
#
# Copyright (C) 2020 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NIT (Número de Identificación Tributaria, El Salvador tax number).

This number consists of 14 digits, usually separated into four groups
using hyphens to make it easier to read, like XXXX-XXXXXX-XXX-X.

The first four digits indicate the code for the municipality of birth
for natural persons or the municipality of stablishment for juridical
persons. NIT for El Salvador nationals begins with either 0 or 1, and
for foreigners it begins with 9.

The following six digits indicate the date of birth for the natural
person, or the stablishment date for the juridical person, using the
format DDMMYY, where DD is the day, MM is the month, and YY is the
year. For example XXXX-051180-XXX-X is (November 5 1980)

The next 3 digits are a sequential number.

The last digit is the check digit, which is used to verify the number
was correctly typed.

More information:

* https://es.wikipedia.org/wiki/Identificaci%C3%B3n_tributaria
* https://www.listasal.info/articulos/nit-el-salvador.shtml
* https://tramitesyrequisitos.com/el-salvador/nit/#Estructura_del_NIT
* https://www.svcommunity.org/forum/programacioacuten/como-calcular-digito-verificador-del-dui-y-nit/msg951882/#msg951882

>>> validate('0614-050707-104-8')
'06140507071048'
>>> validate('SV 0614-050707-104-8')
'06140507071048'
>>> validate('0614-050707-104-0')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('12345678')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('06140507071048')
'0614-050707-104-8'
"""  # noqa: E501

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace.
    """
    number = clean(number, ' -').upper().strip()
    if number.startswith('SV'):
        return number[2:]
    return number


def calc_check_digit(number):
    """Calculate the check digit."""
    # Old NIT
    if number[10:13] <= '100':
        weights = (14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2)
        total = sum(int(n) * w for n, w in zip(number, weights))
        return str((total % 11) % 10)
    # New NIT
    weights = (2, 7, 6, 5, 4, 3, 2, 7, 6, 5, 4, 3, 2)
    total = sum(int(n) * w for n, w in zip(number, weights))
    return str((-total % 11) % 10)


def validate(number):
    """Check if the number is a valid El Salvador NIT number.

    This checks the length, formatting and check digit.
    """
    number = compact(number)
    if len(number) != 14:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if number[0] not in ('0', '1', '9'):
        raise InvalidComponent()
    if number[-1] != calc_check_digit(number):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid El Salvador NIT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '-'.join([number[:4], number[4:-4], number[-4:-1], number[-1]])



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Final

GEN_AI_COMPLETION: Final = "gen_ai.completion"
"""
Deprecated: Removed, no replacement at this time.
"""

GEN_AI_OPENAI_REQUEST_RESPONSE_FORMAT: Final = (
    "gen_ai.openai.request.response_format"
)
"""
The response format that is requested.
"""

GEN_AI_OPENAI_REQUEST_SEED: Final = "gen_ai.openai.request.seed"
"""
Requests with same seed value more likely to return same result.
"""

GEN_AI_OPENAI_REQUEST_SERVICE_TIER: Final = (
    "gen_ai.openai.request.service_tier"
)
"""
The service tier requested. May be a specific tier, default, or auto.
"""

GEN_AI_OPENAI_RESPONSE_SERVICE_TIER: Final = (
    "gen_ai.openai.response.service_tier"
)
"""
The service tier used for the response.
"""

GEN_AI_OPENAI_RESPONSE_SYSTEM_FINGERPRINT: Final = (
    "gen_ai.openai.response.system_fingerprint"
)
"""
A fingerprint to track any eventual change in the Generative AI environment.
"""

GEN_AI_OPERATION_NAME: Final = "gen_ai.operation.name"
"""
The name of the operation being performed.
Note: If one of the predefined values applies, but specific system uses a different name it's RECOMMENDED to document it in the semantic conventions for specific GenAI system and use system-specific name in the instrumentation. If a different name is not documented, instrumentation libraries SHOULD use applicable predefined value.
"""

GEN_AI_PROMPT: Final = "gen_ai.prompt"
"""
Deprecated: Removed, no replacement at this time.
"""

GEN_AI_REQUEST_ENCODING_FORMATS: Final = "gen_ai.request.encoding_formats"
"""
The encoding formats requested in an embeddings operation, if specified.
Note: In some GenAI systems the encoding formats are called embedding types. Also, some GenAI systems only accept a single format per request.
"""

GEN_AI_REQUEST_FREQUENCY_PENALTY: Final = "gen_ai.request.frequency_penalty"
"""
The frequency penalty setting for the GenAI request.
"""

GEN_AI_REQUEST_MAX_TOKENS: Final = "gen_ai.request.max_tokens"
"""
The maximum number of tokens the model generates for a request.
"""

GEN_AI_REQUEST_MODEL: Final = "gen_ai.request.model"
"""
The name of the GenAI model a request is being made to.
"""

GEN_AI_REQUEST_PRESENCE_PENALTY: Final = "gen_ai.request.presence_penalty"
"""
The presence penalty setting for the GenAI request.
"""

GEN_AI_REQUEST_STOP_SEQUENCES: Final = "gen_ai.request.stop_sequences"
"""
List of sequences that the model will use to stop generating further tokens.
"""

GEN_AI_REQUEST_TEMPERATURE: Final = "gen_ai.request.temperature"
"""
The temperature setting for the GenAI request.
"""

GEN_AI_REQUEST_TOP_K: Final = "gen_ai.request.top_k"
"""
The top_k sampling setting for the GenAI request.
"""

GEN_AI_REQUEST_TOP_P: Final = "gen_ai.request.top_p"
"""
The top_p sampling setting for the GenAI request.
"""

GEN_AI_RESPONSE_FINISH_REASONS: Final = "gen_ai.response.finish_reasons"
"""
Array of reasons the model stopped generating tokens, corresponding to each generation received.
"""

GEN_AI_RESPONSE_ID: Final = "gen_ai.response.id"
"""
The unique identifier for the completion.
"""

GEN_AI_RESPONSE_MODEL: Final = "gen_ai.response.model"
"""
The name of the model that generated the response.
"""

GEN_AI_SYSTEM: Final = "gen_ai.system"
"""
The Generative AI product as identified by the client or server instrumentation.
Note: The `gen_ai.system` describes a family of GenAI models with specific model identified
by `gen_ai.request.model` and `gen_ai.response.model` attributes.

The actual GenAI product may differ from the one identified by the client.
For example, when using OpenAI client libraries to communicate with Mistral, the `gen_ai.system`
is set to `openai` based on the instrumentation's best knowledge.

For custom model, a custom friendly name SHOULD be used.
If none of these options apply, the `gen_ai.system` SHOULD be set to `_OTHER`.
"""

GEN_AI_TOKEN_TYPE: Final = "gen_ai.token.type"
"""
The type of token being counted.
"""

GEN_AI_USAGE_COMPLETION_TOKENS: Final = "gen_ai.usage.completion_tokens"
"""
Deprecated: Replaced by `gen_ai.usage.output_tokens` attribute.
"""

GEN_AI_USAGE_INPUT_TOKENS: Final = "gen_ai.usage.input_tokens"
"""
The number of tokens used in the GenAI input (prompt).
"""

GEN_AI_USAGE_OUTPUT_TOKENS: Final = "gen_ai.usage.output_tokens"
"""
The number of tokens used in the GenAI response (completion).
"""

GEN_AI_USAGE_PROMPT_TOKENS: Final = "gen_ai.usage.prompt_tokens"
"""
Deprecated: Replaced by `gen_ai.usage.input_tokens` attribute.
"""


class GenAiOpenaiRequestResponseFormatValues(Enum):
    TEXT = "text"
    """Text response format."""
    JSON_OBJECT = "json_object"
    """JSON object response format."""
    JSON_SCHEMA = "json_schema"
    """JSON schema response format."""


class GenAiOpenaiRequestServiceTierValues(Enum):
    AUTO = "auto"
    """The system will utilize scale tier credits until they are exhausted."""
    DEFAULT = "default"
    """The system will utilize the default scale tier."""


class GenAiOperationNameValues(Enum):
    CHAT = "chat"
    """Chat completion operation such as [OpenAI Chat API](https://platform.openai.com/docs/api-reference/chat)."""
    TEXT_COMPLETION = "text_completion"
    """Text completions operation such as [OpenAI Completions API (Legacy)](https://platform.openai.com/docs/api-reference/completions)."""
    EMBEDDINGS = "embeddings"
    """Embeddings operation such as [OpenAI Create embeddings API](https://platform.openai.com/docs/api-reference/embeddings/create)."""


class GenAiSystemValues(Enum):
    OPENAI = "openai"
    """OpenAI."""
    VERTEX_AI = "vertex_ai"
    """Vertex AI."""
    ANTHROPIC = "anthropic"
    """Anthropic."""
    COHERE = "cohere"
    """Cohere."""
    AZ_AI_INFERENCE = "az.ai.inference"
    """Azure AI Inference."""
    IBM_WATSONX_AI = "ibm.watsonx.ai"
    """IBM Watsonx AI."""
    AWS_BEDROCK = "aws.bedrock"
    """AWS Bedrock."""


class GenAiTokenTypeValues(Enum):
    INPUT = "input"
    """Input tokens (prompt, input, etc.)."""
    COMPLETION = "output"
    """Output tokens (completion, response, etc.)."""



from product_api.dto import Part
from typing import (
    Dict,
    List,
    Optional,
)


# noinspection PyShadowingBuiltins
class SaleLine(object):
    def __init__(
        self,
        line_number,                    # type: int
        part,                           # type: Part
        level,                          # type: int
        item_id,                        # type: str
        item_price,                     # type: float
        quantity,                       # type: float
        default_qty,                    # type: int
        multiplied_qty=None,            # type: float
        added_qty=None,                 # type: Optional[float]
        overwritten_unit_price=None,    # type: Optional[float]
        comment=None,                   # type: Optional[str]
        service_period=None,            # type: Optional[str]
        seat=None,                      # type: Optional[int]
        custom_properties=None,         # type: Dict[str, str]
        discounts_applied=None,         # type: Optional[str]
    ):
        # type: (...) -> None

        self.line_number = line_number
        self.part = part
        self.level = level
        self.item_id = item_id
        self.item_price = item_price
        self.overwritten_unit_price = overwritten_unit_price
        self.quantity = quantity
        self.added_qty = added_qty
        self.default_qty = default_qty
        self.multiplied_qty = multiplied_qty
        self.comment = comment
        self.service_period = service_period
        self.seat = seat
        self.custom_properties = custom_properties
        self.discounts_applied = discounts_applied

        self.fraction_parent = None     # type: Optional[int]
        self.parent = None              # type: Optional[SaleLine]
        self.sons = None                # type: Optional[List[SaleLine]]

    def get_custom_property(
        self,   # type: SaleLine
        key,    # type: str
    ):
        # type: (...) -> Optional[str]

        return self.custom_properties.get(key, None)

    def get_line_id(
        self,   # type: SaleLine
    ):
        # type: (...) -> str

        return "{}.{}-{}.{}".format(
            self.line_number,
            self.level,
            self.item_id,
            self.part.part_code,
        )



from typing import Dict, Union, Optional

from ._DataType import DataType
from ._DefaultToken import DefaultToken
from ._Message import Message


class HttpResponseMessage(Message):
    def __init__(self, status_code, body=None, headers=None):
        # type: (int, Optional[Union[str, bytes]], Dict[str, str]) -> None
        self.status_code = status_code
        self.body = body
        self.http_headers = headers if headers is not None else {}

        data = str(self.status_code).encode("utf-8") + b"\0" + str(len(self.http_headers)).encode("utf-8") + b"\0"
        for header in self.http_headers.keys():
            data += header.encode("utf-8") + b"\0" + self.http_headers[header].encode("utf-8") + b"\0"

        if body is None:
            data += b""
        else:
            body = self.body
            if isinstance(body, str):
                body = body.encode("utf-8")
            data += body

        super(HttpResponseMessage, self).__init__(DefaultToken.TK_HTTP_MSG, DataType.http_response, data)



import persistence

from persistence import Driver

_MWPOS_BASE_HMAC = 0x18122008


class UserHelper:
    def __init__(self, mbcontext):
        # type: (msgbus.MBEasyContext) -> None
        self.mbcontext = mbcontext
        self.conn = None  # type: persistence.Connection

    def __enter__(self):
        self.conn = Driver().open(self.mbcontext)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.conn:
            self.conn.close()
        return exc_val

    def get_all_users(self):
        # type: () -> dict
        return {x.get_entry(0): (x.get_entry(1), x.get_entry(2), x.get_entry(3), x.get_entry(4))
                for x in self.conn.select("""select UserId, LongName, Level, Status from Users""")}

    def get_all_logged_users(self):
        query = """select operatorid
                     from posctrl.UserSession
                    where closetime IS NULL"""

        select = self.conn.select(query)
        return {x.get_entry(0) for x in select}



# luhn.py - functions for performing the Luhn and Luhn mod N algorithms
#
# Copyright (C) 2010-2021 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""The Luhn and Luhn mod N algorithms.

The Luhn algorithm is used to detect most accidental errors in various
identification numbers.

>>> validate('7894')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> checksum('7894')
6
>>> calc_check_digit('7894')
'9'
>>> validate('78949')
'78949'

An alternative alphabet can be provided to use the Luhn mod N algorithm.
The default alphabet is '0123456789'.

>>> validate('1234', alphabet='0123456789abcdef')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> checksum('1234', alphabet='0123456789abcdef')
14
"""

from stdnum.exceptions import *


def checksum(number, alphabet='0123456789'):
    """Calculate the Luhn checksum over the provided number. The checksum
    is returned as an int. Valid numbers should have a checksum of 0."""
    n = len(alphabet)
    number = tuple(alphabet.index(i)
                   for i in reversed(str(number)))
    return (sum(number[::2]) +
            sum(sum(divmod(i * 2, n))
                for i in number[1::2])) % n


def validate(number, alphabet='0123456789'):
    """Check if the number provided passes the Luhn checksum."""
    if not bool(number):
        raise InvalidFormat()
    try:
        valid = checksum(number, alphabet) == 0
    except Exception:  # noqa: B902
        raise InvalidFormat()
    if not valid:
        raise InvalidChecksum()
    return number


def is_valid(number, alphabet='0123456789'):
    """Check if the number passes the Luhn checksum."""
    try:
        return bool(validate(number, alphabet))
    except ValidationError:
        return False


def calc_check_digit(number, alphabet='0123456789'):
    """Calculate the extra digit that should be appended to the number to
    make it a valid number."""
    ck = checksum(str(number) + alphabet[0], alphabet)
    return alphabet[-ck]



from ._TableError import TableError
from typing import Dict


class TabToJoinOpenedInAnotherPOS(TableError):

    def __init__(
        self,       # type: TabToJoinOpenedInAnotherPOS
        tab_name,   # type: str
    ):
        # type: (...) -> None

        message = "$TAB_TO_JOIN_IS_OPENED_IN_ANOTHER_POS|{}".format(tab_name)
        super(TabToJoinOpenedInAnotherPOS, self).__init__(message)

        self.tab_name = tab_name

    def get_extra_data(
        self,   # type: TabToJoinOpenedInAnotherPOS
    ):
        # type: (...) -> Dict[str, str]

        return {
            "tabName": self.tab_name,
        }



from datetime import datetime, timedelta
from logging import Logger
from typing import List

from production.box._OrderChangerProductionBox import OrderChangerProductionBox
from production.command import (
    ProductionCommandProcessor,
    PurgeOrdersCommand,
)
from production.model import ProductionOrder


class RemoveReprocessedOrdersBox(OrderChangerProductionBox):
    def __init__(
        self,
        name,                       # type: str
        sons,                       # type: List[str]
        command_processor,          # type: ProductionCommandProcessor
        reprocessed_time_limit,     # type: int
        logger=None,                # type: Logger
    ):
        # type: (...) -> None

        super(RemoveReprocessedOrdersBox, self).__init__(name, sons, logger)
        self.command_processor = command_processor
        self.reprocessed_time_limit = int(reprocessed_time_limit)
        self.logger = logger

    def change_order(self, order):
        if order.first_processing and not order.purged:
            order_creation_time = datetime.strptime(order.created_at[0:19], "%Y-%m-%dT%H:%M:%S")
            if (datetime.now() - order_creation_time) > timedelta(seconds=self.reprocessed_time_limit):
                self.logger.error("Order removed after reprocessing time limit. Order: \n{}".format(str(order)))
                self._purge_order(order=order)
                return None

        return order

    def _purge_order(
        self,
        order,  # type: ProductionOrder
    ):
        # type: (...) -> None

        self.command_processor.handle_purge_orders_command(command=PurgeOrdersCommand([order.order_id]))



# -*- coding: utf-8 -*-

import os
import json
from xml.etree import cElementTree as eTree

import cfgtools
from posot import OrderTaker
from sysactions import (
    show_messagebox,
    show_keyboard,
    StopAction,
)

import helper
from mw_helper import show_message_options_dialog
from typing import (
    Optional,
    Union
)

from api.order import get_order_service
from utilfunctions import (
    validate_document,
    update_custom_properties,
    get_custom_property_value,
    input_document_pinpad_dialog
)
from model.customexception import RequiredDocumentException
from order_api import OrderService
from order_api.dto import CustomProperty

document_required_value = None


def _get_order_loyalty_customer_attribute(
    pos_id,                 # type: str
    order_picture,          # type: eTree.Element
    attribute="document",   # type: Optional[str]
):
    # type: (...) -> str

    from actions.loyalty import get_table_loyalty_customer_attribute

    loyalty_info = order_picture.find(".//OrderProperty[@key='LOYALTY_CUSTOMER_INFO']")
    table_id = order_picture.find(".//OrderProperty[@key='TABLE_ID']")
    if loyalty_info is not None:
        return json.loads(loyalty_info.get('value')).get('customerInfo', {}).get('customer', {}).get(attribute, '')
    elif table_id is not None:
        return get_table_loyalty_customer_attribute(
            pos_id=pos_id,
            table_id=table_id.get('value'),
            attribute=attribute,
        )
    return ''


def _set_document_order_property(
    order_service,  # type: OrderService
    document,       # type: str
    order_id,       # type: str
):
    # type: (...) -> None

    order_properties_dict = update_custom_properties(customer_doc=document)
    if order_properties_dict:
        order_service.set_order_custom_properties(
            custom_properties=[
                CustomProperty(
                    key=key,
                    value=value,
                ) for key, value in order_properties_dict.items()
            ],
            order_id=int(order_id),
        )


def validate_required_document(
    pos_id,                         # type: str
    pos_ot,                         # type: OrderTaker
    order,                          # type: eTree.Element
    ask_document_pinpad=False,      # type: Optional[bool]
):
    # type: (...) -> Union[str, bool]

    total_order = float(order.get("totalGross")) + float(order.get("tip", 0))
    required_value = get_document_required_value()
    ask_for_doc = required_value > 0.0 and (total_order >= required_value)

    document = get_custom_property_value(
        pos_id=pos_id,
        custom_property='CUSTOMER_DOC',
        order_picture=order,
    )
    loyalty_document = _get_order_loyalty_customer_attribute(
        pos_id=pos_id,
        order_picture=order,
    )
    order_service = get_order_service(pos_id=pos_id)
    if str(loyalty_document) == document:
        buttons = "$CONFIRM|$INFORM_ANOTHER|$WITHOUT_DOCUMENT"
        message = "$CONFIRM_CPF_CNPJ|{0}".format(helper.format_cpf_cnpj(value=loyalty_document))
        resp = show_message_options_dialog(
            pos_id=pos_id,
            buttons=buttons,
            title="$CONFIRMATION",
            message=message,
        )
        if resp == 2 or resp is None:
            return False

        elif resp == 1:
            loyalty_document = get_customer_doc(
                pos_id=pos_id,
                pos_ot=pos_ot,
                ask_document_pinpad=ask_document_pinpad,
            )
        _set_document_order_property(
            order_service=order_service,
            document=loyalty_document,
            order_id=order.get("orderId"),
        )
        return True

    elif ask_for_doc and document is None:
        message = "$DOC_REQUIRED|%s" % "{0:.2f}".format(required_value)
        show_messagebox(pos_id, message, "$INFORMATION")

        document = get_customer_doc(
            pos_id=pos_id,
            pos_ot=pos_ot,
            required=True,
            ask_document_pinpad=ask_document_pinpad,
        )
        _set_document_order_property(
            order_service=order_service,
            document=document,
            order_id=order.get("orderId"),
        )
        return document

    return False


def get_customer_doc(
    pos_id,                      # type: str
    pos_ot=None,                 # type: Optional[OrderTaker]
    default_doc="",              # type: Optional[str]
    required=False,              # type: bool
    ask_document_pinpad=False,   # type: Optional[bool]
    stop_action=True,            # type: Optional[bool]
    show_cancel_button=True,     # type: Optional[bool]
):
    # type: (...) -> Optional[str]

    valid = False
    customer_doc = None
    while not valid:
        if ask_document_pinpad:
            customer_doc = input_document_pinpad_dialog(pos_id)

        if customer_doc is None:
            message = "$ONLY_NUMBERS"
            title = "$ENTER_CUSTOMER_DOCUMENT"
            previous_button = "|$PREVIOUS" if show_cancel_button else ""
            if required:
                buttons = "$OK" + previous_button
            else:
                buttons = "$OK|$NOT_INFORM" + previous_button

            customer_doc = show_keyboard(
                posid=pos_id,
                message=message,
                title=title,
                buttons=buttons,
                mask='CPF_CNPJ',
                numpad=True,
                defvalue=default_doc,
            )

        if customer_doc is None:
            if required:
                raise RequiredDocumentException()
            if stop_action:
                raise StopAction()
            return None

        if customer_doc in ("NOT_INFORM", "EMPTY_DOC") and not required:
            customer_doc = ""
            break

        valid = validate_document(customer_doc)
        if not valid:
            customer_doc = None
            show_messagebox(pos_id, "$INVALID_CPF_CNPJ")

    if pos_ot is not None:
        fill_additional_info(
            pos_ot=pos_ot,
            properties="CPF",
            value=customer_doc,
        )

    return customer_doc


def get_customer_phone_number(
    pos_id,                       # type: str
    required=False,               # type: bool
    stop_action=True,             # type: Optional[bool]
):
    # type: (...) -> Optional[str]

    customer_phone = None
    while True:

        message = "$ONLY_NUMBERS"
        customer_phone = show_keyboard(
            posid=pos_id,
            message=message,
            title="$ENTER_CLIENT_PHONE",
            buttons="$OK|$NOT_INFORM|$PREVIOUS",
            mask="PHONE_NUMBER",
            numpad=True,
        )

        if customer_phone in (None, "NOT_INFORM"):
            return _get_not_informed_customer_phone_input(
                customer_phone=customer_phone,
                required=required,
                stop_action=stop_action,
            )

        if customer_phone == "" and not required:
            break

        if len(customer_phone) not in (10, 11):
            customer_phone = None
            show_messagebox(
                posid=pos_id,
                message="$INVALID_CUSTOMER_PHONE",
            )
        else:
            break

        show_messagebox(
            posid=pos_id,
            message="$INVALID_CUSTOMER_PHONE",
        )

    return customer_phone


def _get_not_informed_customer_phone_input(
    customer_phone,     # type: Optional[str]
    required,           # type: bool
    stop_action,        # type: bool
):
    # type: (...) -> Optional[str]

    if required:
        raise RequiredDocumentException()

    if stop_action:
        raise StopAction()

    return customer_phone


def fill_additional_info(
        pos_ot,         # type: OrderTaker
        properties,     # type: str
        value           # type: str
):
    # type: (...) -> None

    if pos_ot.additionalInfo:
        pos_ot.additionalInfo += "|"
    else:
        pos_ot.additionalInfo = ""

    pos_ot.additionalInfo += "{0}={1}".format(properties, value)


def get_document_required_value():
    # type: (...) -> float

    global document_required_value

    if document_required_value is None:
        config = cfgtools.read(os.environ["LOADERCFG"])
        document_required_value = float(config.find_value("Customizations.DocumentRequiredValue") or 0)

    return document_required_value



# bankaccount.py - functions for handling New Zealand bank account numbers
# coding: utf-8
#
# Copyright (C) 2019 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""New Zealand bank account number

The New Zealand bank account numbers consist of 16 digits. The first two
represent the bank, followed by four for the branch, seven digits for the
account base number and three for the account type.

More information:

* https://en.wikipedia.org/wiki/New_Zealand_bank_account_number

>>> validate('01-0242-0100194-00')
'0102420100194000'
>>> validate('01-0242-0100195-00')  # invalid check digits
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('01-9999-0100197-00')  # invalid branch
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> format('0102420100194000')
'01-0242-0100194-000'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


# The following algorithms and weights were taken from:
# https://www.ird.govt.nz/software-providers/explore-products-contents/reporting/withholding-taxes/rwt-and-nrwt-certificate-filing-options.html#02
# with the modification to use max 7 digits for the account base
# instead of 8 (and leaving out algorithm C).

# The algorithm to choose based on the bank
_algorithms = {
    '01': 'A', '02': 'A', '03': 'A', '04': 'A', '06': 'A', '08': 'D',
    '09': 'E', '10': 'A', '11': 'A', '12': 'A', '13': 'A', '14': 'A',
    '15': 'A', '16': 'A', '17': 'A', '18': 'A', '19': 'A', '20': 'A',
    '21': 'A', '22': 'A', '23': 'A', '24': 'A', '25': 'F', '26': 'G',
    '27': 'A', '28': 'G', '29': 'G', '30': 'A', '31': 'X', '33': 'F',
    '35': 'A', '38': 'A',
}

# The different weights for the different checksum algorithms
_weights = {
    'A': (0, 0, 6, 3, 7, 9, 0, 10, 5, 8, 4, 2, 1, 0, 0, 0),
    'B': (0, 0, 0, 0, 0, 0, 0, 10, 5, 8, 4, 2, 1, 0, 0, 0),
    'D': (0, 0, 0, 0, 0, 0, 7, 6, 5, 4, 3, 2, 1, 0, 0, 0),
    'E': (0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 4, 3, 2, 0, 0, 1),
    'F': (0, 0, 0, 0, 0, 0, 1, 7, 3, 1, 7, 3, 1, 0, 0, 0),
    'G': (0, 0, 0, 0, 0, 0, 1, 3, 7, 1, 3, 7, 1, 3, 7, 1),
    'X': (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
}

# The moduli to use per algorithm
_moduli = {
    'A': (11, 11),
    'B': (11, 11),
    'D': (11, 11),
    'E': (9, 11),
    'F': (10, 10),
    'G': (9, 10),
    'X': (1, 1),
}


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    number = clean(number).strip().replace(' ', '-').split('-')
    if len(number) == 4:
        # zero pad the different sections if they are found
        lengths = (2, 4, 7, 3)
        return ''.join(n.zfill(l) for n, l in zip(number, lengths))
    else:
        # otherwise zero pad the account type
        number = ''.join(number)
        return number[:13] + number[13:].zfill(3)


def info(number):
    """Return a dictionary of data about the supplied number. This typically
    returns the name of the bank and branch and a BIC if it is valid."""
    number = compact(number)
    from stdnum import numdb
    info = {}
    for _nr, found in numdb.get('nz/banks').info(number):
        info.update(found)
    return info


def _calc_checksum(number):
    # pick the algorithm and parameters
    algorithm = _algorithms.get(number[:2], 'X')
    if algorithm == 'A' and number[6:13] >= '0990000':
        algorithm = 'B'
    weights = _weights[algorithm]
    mod1, mod2 = _moduli[algorithm]
    # calculate the checksum
    return sum(
        c % mod1 if c > mod1 else c for c in
        (w * int(n) for w, n in zip(weights, number))) % mod2


def validate(number):
    """Check if the number provided is a valid bank account number."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 16:
        raise InvalidLength()
    if _calc_checksum(number) != 0:
        raise InvalidChecksum()
    i = info(number)
    if 'bank' not in i or 'branch' not in i:
        raise InvalidComponent()
    return number


def is_valid(number):
    """Check if the number provided is a valid bank account number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '-'.join([
        number[:2],
        number[2:6],
        number[6:13],
        number[13:]])



"""Initial version

Revision ID: 5bdc013c2a92
Revises:
Create Date: 2023-10-13 10:29:57.476579

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy import (
    Column,
    TEXT,
    INTEGER,
    BLOB,
    DATETIME,
    ForeignKey,
)

# revision identifiers, used by Alembic.
revision: str = '5bdc013c2a92'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    pending_event_table = op.create_table(
        "PendingEvent",
        Column("Id", TEXT, nullable=False, primary_key=True),
        Column("CreatedAt", DATETIME, nullable=False),
        Column("ExternalCode", TEXT, nullable=True),
        Column("Subject", TEXT, nullable=False),
        Column("EventType", TEXT),
        Column("Data", BLOB),
        Column("Headers", TEXT),
        Column("AckedAt", DATETIME),
        Column("NextRetryAt", DATETIME),
        Column("RetryIndex", INTEGER),
        Column("AbandonedAt", DATETIME),
        Column("Target", TEXT),
        Column("AsyncReplySubject", TEXT),
        Column("AsyncIdField", TEXT),
        Column("AsyncId", TEXT),
    )

    op.create_index(
        index_name="PendingEvent_AsyncReplySubject_AsyncId",
        table_name="PendingEvent",
        columns=("AsyncReplySubject", "AsyncId"),
    )

    op.create_index(
        index_name="PendingEvent_ExternalCode",
        table_name="PendingEvent",
        columns=("ExternalCode",),
        unique=True,
    )

    op.create_index(
        index_name="PendingEvent_CreatedAt",
        table_name="PendingEvent",
        columns=("CreatedAt",),
    )

    op.create_table(
        "PublicationHistory",
        Column("DefinitionId", TEXT, ForeignKey(pending_event_table.c.Id), nullable=False),
        Column("CreatedAt", DATETIME, nullable=False),
        Column("Action", TEXT, nullable=False),
    )

    op.create_index(
        index_name="PublicationHistory_DefinitionId",
        table_name="PublicationHistory",
        columns=("DefinitionId", ),
    )


def downgrade() -> None:
    op.drop_table("PendingEvent")
    op.drop_table("PublicationHistory")
